import sys

from oasys2.widget.widget import OWWidget
from orangewidget import gui
from orangewidget.settings import Setting

from AnyQt.QtWidgets import QApplication
from AnyQt.QtCore import QRect

import oasys2.widget.gui as oasysgui
from oasys2.widget.gui import ConfirmDialog, MessageDialog

class AutomaticElement(OWWidget):
    want_main_area = 1
    is_automatic_run = Setting(True)

    MAX_WIDTH          = 1320
    MAX_HEIGHT         = 720
    CONTROL_AREA_WIDTH = 405
    TABS_AREA_HEIGHT   = 560

    def __init__(self, show_automatic_box=True):
        super().__init__()

        geom = QApplication.primaryScreen().geometry()
        self.setGeometry(QRect(round(geom.width()*0.05),
                               round(geom.height()*0.05),
                               round(min(geom.width()*0.98, self.MAX_WIDTH)),
                               round(min(geom.height()*0.95, self.MAX_HEIGHT))))
        self.setMaximumHeight(self.geometry().height())
        self.setMaximumWidth(self.geometry().width())
        self.controlArea.setFixedWidth(self.CONTROL_AREA_WIDTH)

        if show_automatic_box:
            self.general_options_box = oasysgui.widgetBox(self.controlArea, "General Options", addSpace=True, orientation="horizontal", width=self.CONTROL_AREA_WIDTH-5)
            gui.checkBox(self.general_options_box, self, 'is_automatic_run', 'Automatic Execution')
            self.TABS_AREA_HEIGHT = 555
        else:
            self.TABS_AREA_HEIGHT = 615

    def call_reset_settings(self):
        if ConfirmDialog.confirmed(parent=self, message="Confirm Reset of the Fields?"):
            try:    self._reset_settings()
            except: pass

    def prompt_exception(self, exception: Exception):
        MessageDialog.message(self, str(exception), "Exception occured in OASYS", "critical")
        if self.IS_DEVELOP: raise exception




__author__ = 'labx'

import os
import random
import sys
import copy

import numpy
import h5py
import re
import time

from scipy import optimize
from numpy import asarray
import scipy.constants as codata


from AnyQt.QtCore import QSettings
from AnyQt.QtWidgets import QWidget, QGridLayout

from matplotlib import rcParams
from matplotlib.patches import FancyArrowPatch, ArrowStyle

from orangewidget.gui import OWComponent

from oasys2.widget import gui
from oasys2.widget.util import congruence
from oasys2.widget.gui import Styles
from oasys2.widget.util.widget_objects import TriggerIn, TriggerOut

from srxraylib.util.histograms import get_sigma, get_fwhm, get_average

from shadow4.beam.s4_beam import S4Beam

from orangecontrib.shadow4.util import materials_library as ml


class ShadowCongruence():
    @classmethod
    def check_empty_beam(cls, input_beam):
        if input_beam is None: return False
        elif not hasattr(input_beam, "rays"): return False
        elif len(input_beam.rays) == 0: return False
        else: return True

    @classmethod
    def check_empty_data(cls, input_data):
        if input_data is None:  return False
        else:                   return cls.check_empty_beam(input_data.beam)

    @classmethod
    def check_good_beam(cls, input_beam):
        return len(input_beam.rays[numpy.where(input_beam.rays[:, 9] == 1)]) > 0

    @classmethod
    def checkBraggFile(cls, file_name):
        file = open(file_name, "r")

        try:
            rows = file.readlines()

            if len(rows) < 10: raise Exception("Bragg file malformed, please check input")

            if "# Bragg version," in rows[0]:
                pass # version 2
            else:
                first_row = ShadowCongruence.__get_numbers(rows[0].strip())

                if not len(first_row) == 3: raise Exception("Bragg file malformed, please check input")

                second_row = ShadowCongruence.__get_numbers(rows[1].strip())
                if not len(second_row) == 3: raise Exception("Bragg file malformed, please check input")

                if not (rows[2].strip().startswith("(") and \
                                rows[3].strip().startswith("(") and \
                                rows[4].strip().startswith("(") and \
                                rows[5].strip().startswith("(")):
                    raise Exception("Bragg file malformed, please check input")

                seventh_row = ShadowCongruence.__get_numbers(rows[6].strip())
                if not len(seventh_row) == 3: raise Exception("Bragg file malformed, please check input")

                eighth_row = ShadowCongruence.__get_numbers(rows[7].strip())
                if not len(eighth_row) == 3: raise Exception("Bragg file malformed, please check input")

                nineth_row = ShadowCongruence.__get_numbers(rows[8].strip())
                if not len(nineth_row) == 1: raise Exception("Bragg file malformed, please check input")
        except Exception as e:
            file.close()

            raise e

    @classmethod
    def checkPreReflFile(cls, file_name):
        file = open(file_name, "r")

        try:
            rows = file.readlines()

            if len(rows) < 3: raise Exception("PreRefl file malformed, please check input")

            first_row = ShadowCongruence.__get_numbers(rows[0].strip())
            if not len(first_row) == 4: raise Exception("PreRefl file malformed, please check input")

            second_row = ShadowCongruence.__get_numbers(rows[1].strip())
            if not len(second_row) == 1: raise Exception("PreRefl file malformed, please check input")

            try:
                elements = int(second_row[0])
            except:
                raise Exception("PreRefl file malformed, please check input")

            if len(rows) != (elements*2) + 2: raise Exception("PreRefl file malformed, please check input")
        except Exception as e:
            file.close()

            raise e

    @classmethod
    def checkPreMLayerFile(cls, file_name):
        file = open(file_name, "r")

        try:
            rows = file.readlines()

            if len(rows) < 2: raise Exception("PreMLayer file malformed, please check input")

            first_row = ShadowCongruence.__get_numbers(rows[0].strip())
            if not len(first_row) == 1: raise Exception("PreMLayer file malformed, please check input")

            try:
                elements = int(first_row[0])
            except:
                raise Exception("PreRefl file malformed, please check input")

            second_row = ShadowCongruence.__get_numbers(rows[1].strip())
            if not len(second_row) == int(elements): raise Exception("PreMLayer file malformed, please check input")

            try:
                separator_row = ShadowCongruence.__get_numbers(rows[2 + elements*3].strip())
                if not len(separator_row) == 1: raise Exception("PreMLayer file malformed, please check input")
            except:
                raise Exception("PreRefl file malformed, please check input")

            next_row = ShadowCongruence.__get_numbers(rows[2 + elements*3 + 1].strip())
            if not len(next_row) == 4: raise Exception("PreMLayer file malformed, please check input")
        except Exception as e:
            file.close()

            raise e

    @classmethod
    def check2ColumnFormatFile(cls, file_name, specific_name):
        try:
            if file_name.startswith('/'):
                values = numpy.loadtxt(os.path.abspath(file_name))
            else:
                values = numpy.loadtxt(os.path.abspath(os.path.join(os.path.curdir, file_name)))
        except:
            raise Exception(specific_name + " file malformed (should be 2 or more columns of numbers, separated by spaces), please check input")

        if len(values) < 2: raise Exception(specific_name + " file malformed (should be 2 or more columns of numbers, separated by spaces), please check input")

    @classmethod
    def checkErrorProfileFile(cls, file_name):
        file = open(file_name, "r")

        try:
            rows = file.readlines()

            if len(rows) < 2: raise Exception("Surface Error file malformed, please check input")

            first_row = ShadowCongruence.__get_numbers(rows[0].strip())
            if not len(first_row) == 2: raise Exception("Surface Error file malformed, please check input")

            n_x = int(first_row[0])

            if n_x > 500:
                raise Exception("Malformed file: maximum allowed point in X direction is 500")

        except Exception as e:
            file.close()

            raise e

    @classmethod
    def __get_numbers(cls, string):
        values = string.strip().split(" ")

        numbers = []
        for value in values:
            if not value == "":
                try:
                    numbers.append(value)
                except:
                    pass

        return numbers

class ShadowStatisticData:
    intensity = 0.0
    total_number_of_rays = 0
    total_good_rays = 0
    total_lost_rays = 0

    def __init__(self, intensity = 0.0,
                 total_number_of_rays = 0,
                 total_good_rays = 0,
                 total_lost_rays = 0):
        self.intensity = intensity
        self.total_number_of_rays = total_number_of_rays
        self.total_good_rays = total_good_rays
        self.total_lost_rays = total_lost_rays

class ShadowHistoData(ShadowStatisticData):
    fwhm = 0.0
    x_fwhm_i = 0.0
    x_fwhm_f = 0.0
    y_fwhm = 0.0

    def __init__(self, intensity = 0.0,
                 total_number_of_rays = 0,
                 total_good_rays = 0,
                 total_lost_rays = 0,
                 fwhm = 0.0,
                 x_fwhm_i = 0.0,
                 x_fwhm_f = 0.0,
                 y_fwhm = 0.0):
        super().__init__(intensity, total_number_of_rays, total_good_rays, total_lost_rays)
        self.fwhm = fwhm
        self.x_fwhm_i = x_fwhm_i
        self.x_fwhm_f = x_fwhm_f
        self.y_fwhm = y_fwhm

class ShadowPlotData(ShadowStatisticData):
    fwhm_h = 0.0
    fwhm_v = 0.0

    def __init__(self, intensity = 0.0,
                 total_number_of_rays = 0,
                 total_good_rays = 0,
                 total_lost_rays = 0,
                 fwhm_h = 0.0,
                 fwhm_v = 0.0):
        super().__init__(intensity, total_number_of_rays, total_good_rays, total_lost_rays)
        self.fwhm_h = fwhm_h
        self.fwhm_v = fwhm_v

try:
    class ShadowPlot:

        _is_conversione_active = True

        #########################################################################################
        #
        # FOR TEMPORARY USE: FIX AN ERROR IN PYMCA.PLOT.IMAGEWIEW
        #
        #########################################################################################


        @classmethod
        def set_conversion_active(cls, is_active=True):
            ShadowPlot._is_conversione_active = is_active

        @classmethod
        def get_conversion_active(cls):
            return ShadowPlot._is_conversione_active

        """Sample code to add 2D dataset saving as text to ImageView."""


        #########################################################################################
        #
        # WIDGET FOR DETAILED PLOT
        #
        #########################################################################################

        class InfoBoxWidget(QWidget, OWComponent):
            intensity_field = ""
            flux_field = ""
            total_rays_field = ""
            total_good_rays_field = ""
            total_lost_rays_field = ""
            fwhm_h_field = ""
            fwhm_v_field = ""
            sigma_h_field = ""
            sigma_v_field = ""
            centroid_h_field = ""
            centroid_v_field = ""

            def __init__(self, x_scale_factor = 1.0, y_scale_factor = 1.0, is_2d=True):
                super(ShadowPlot.InfoBoxWidget, self).__init__()

                info_box_inner=gui.widgetBox(self, "Info")
                info_box_inner.setFixedHeight(int(518*y_scale_factor))
                info_box_inner.setFixedWidth(int(230*x_scale_factor))

                self.flux_box = gui.widgetBox(info_box_inner, "", addSpace=False, orientation="horizontal")
                self.flux     = gui.lineEdit(self.flux_box, self, "flux_field", "\u03a6 [ph/s/0.1%BW]", tooltip="Flux", labelWidth=115, valueType=str, orientation="horizontal")
                self.flux_box.setVisible(False)

                self.intensity = gui.lineEdit(info_box_inner, self, "intensity_field", "Intensity", tooltip="Intensity", labelWidth=115, valueType=str, orientation="horizontal")
                self.total_rays = gui.lineEdit(info_box_inner, self, "total_rays_field", "Total Rays", tooltip="Total Rays", labelWidth=115, valueType=str, orientation="horizontal")
                self.total_good_rays = gui.lineEdit(info_box_inner, self, "total_good_rays_field", "Total Good Rays", tooltip="Total Good Rays", labelWidth=115, valueType=str, orientation="horizontal")
                self.total_lost_rays = gui.lineEdit(info_box_inner, self, "total_lost_rays_field", "Total Lost Rays", tooltip="Total Lost Rays", labelWidth=115, valueType=str, orientation="horizontal")

                label_h_style = "color: blue"
                label_v_style = "color: red"

                label_box_1 = gui.widgetBox(info_box_inner, "", addSpace=False, orientation="horizontal")
                self.label_h = gui.widgetLabel(label_box_1, "FWHM", labelWidth=115)
                self.label_h.setStyleSheet(label_h_style)
                label_box_1.layout().addWidget(self.label_h)
                self.fwhm_h = gui.lineEdit(label_box_1, self, "fwhm_h_field", "", tooltip="FWHM", labelWidth=115, valueType=str, orientation="horizontal")

                if is_2d:
                    label_box_2 = gui.widgetBox(info_box_inner, "", addSpace=False, orientation="horizontal")
                    self.label_v = gui.widgetLabel(label_box_2, "FWHM", labelWidth=115)
                    self.label_v.setStyleSheet(label_v_style)
                    label_box_2.layout().addWidget(self.label_v)
                    self.fwhm_v = gui.lineEdit(label_box_2, self, "fwhm_v_field", "", tooltip="FWHM", labelWidth=115, valueType=str, orientation="horizontal")

                label_box_1 = gui.widgetBox(info_box_inner, "", addSpace=False, orientation="horizontal")
                self.label_s_h = gui.widgetLabel(label_box_1, "\u03c3", labelWidth=115)
                self.label_s_h.setStyleSheet(label_h_style)
                label_box_1.layout().addWidget(self.label_s_h)
                self.sigma_h = gui.lineEdit(label_box_1, self, "sigma_h_field", "", tooltip="Sigma", labelWidth=115, valueType=str, orientation="horizontal")

                if is_2d:
                    label_box_2 = gui.widgetBox(info_box_inner, "", addSpace=False, orientation="horizontal")
                    self.label_s_v = gui.widgetLabel(label_box_2, "\u03c3", labelWidth=115)
                    self.label_s_v.setStyleSheet(label_v_style)
                    label_box_2.layout().addWidget(self.label_s_v)
                    self.sigma_v = gui.lineEdit(label_box_2, self, "sigma_v_field", "", tooltip="Sigma", labelWidth=115, valueType=str, orientation="horizontal")

                label_box_1 = gui.widgetBox(info_box_inner, "", addSpace=False, orientation="horizontal")
                self.label_c_h = gui.widgetLabel(label_box_1, "centroid", labelWidth=115)
                self.label_c_h.setStyleSheet(label_h_style)
                label_box_1.layout().addWidget(self.label_c_h)
                self.centroid_h = gui.lineEdit(label_box_1, self, "centroid_h_field", "", tooltip="Centroid", labelWidth=115, valueType=str, orientation="horizontal")

                if is_2d:
                    label_box_2 = gui.widgetBox(info_box_inner, "", addSpace=False, orientation="horizontal")
                    self.label_c_v = gui.widgetLabel(label_box_2, "centroid", labelWidth=115)
                    self.label_c_v.setStyleSheet(label_v_style)
                    label_box_2.layout().addWidget(self.label_c_v)
                    self.centroid_v = gui.lineEdit(label_box_2, self, "centroid_v_field", "", tooltip="Sigma", labelWidth=115, valueType=str, orientation="horizontal")

                read_only_style = Styles.line_edit_read_only

                self.intensity.setReadOnly(True)
                self.intensity.setStyleSheet(read_only_style)
                self.flux.setReadOnly(True)
                self.flux.setStyleSheet(read_only_style)
                self.total_rays.setReadOnly(True)
                self.total_rays.setStyleSheet(read_only_style)
                self.total_good_rays.setReadOnly(True)
                self.total_good_rays.setStyleSheet(read_only_style)
                self.total_lost_rays.setReadOnly(True)
                self.total_lost_rays.setStyleSheet(read_only_style)
                self.fwhm_h.setReadOnly(True)
                self.fwhm_h.setStyleSheet(read_only_style)
                self.sigma_h.setReadOnly(True)
                self.sigma_h.setStyleSheet(read_only_style)
                self.centroid_h.setReadOnly(True)
                self.centroid_h.setStyleSheet(read_only_style)

                if is_2d:
                    self.fwhm_v.setReadOnly(True)
                    self.fwhm_v.setStyleSheet(read_only_style)
                    self.sigma_v.setReadOnly(True)
                    self.sigma_v.setStyleSheet(read_only_style)
                    self.centroid_v.setReadOnly(True)
                    self.centroid_v.setStyleSheet(read_only_style)

            def set_flux(self, flux=None):
                if flux is None:
                    self.flux.setText("0.0")
                    self.flux_box.setVisible(False)
                else:
                    self.flux.setText('%.3E' % flux)
                    self.flux_box.setVisible(True)

            def clear(self):
                self.intensity.setText("0.0")
                self.flux.setText("0.0")
                self.flux_box.setVisible(False)
                self.total_rays.setText("0")
                self.total_good_rays.setText("0")
                self.total_lost_rays.setText("0")
                self.fwhm_h.setText("0.0000")
                if hasattr(self, "fwhm_v"):  self.fwhm_v.setText("0.0000")
                self.sigma_h.setText("0.0000")
                if hasattr(self, "sigma_v"):  self.sigma_v.setText("0.0000")
                self.centroid_h.setText("0.0000")
                if hasattr(self, "centroid_v"):  self.centroid_v.setText("0.0000")
                self.boundary_h.setText("")
                if hasattr(self, "boundary_v"):  self.centroid_v.setText("")

        class DetailedHistoWidget(QWidget, OWComponent):

            def __init__(self, x_scale_factor = 1.0, y_scale_factor = 1.0):
                super(ShadowPlot.DetailedHistoWidget, self).__init__()

                self.plot_canvas = gui.plotWindow(roi=False, control=False, position=True, logScale=True, fit=True)
                self.plot_canvas.setDefaultPlotLines(True)
                self.plot_canvas.setActiveCurveStyle(color='blue')
                self.plot_canvas.setMinimumWidth(int(590 * x_scale_factor))
                self.plot_canvas.setMaximumWidth(int(590 * x_scale_factor))

                self.info_box = ShadowPlot.InfoBoxWidget(x_scale_factor, y_scale_factor, is_2d=False)

                layout = QGridLayout()

                layout.addWidget(   self.info_box, 0, 1, 1, 1)
                layout.addWidget(self.plot_canvas, 0, 0, 1, 1)

                layout.setColumnMinimumWidth(0, int(600 * x_scale_factor))
                layout.setColumnMinimumWidth(1, int(230 * x_scale_factor))

                self.setLayout(layout)

            def plot_histo(self, beam, col, nolost, xrange, ref, title, xtitle, ytitle, nbins = 100, xum="", ticket_to_add=None, flux=None):
                ticket = beam.histo1(col, nbins=nbins, xrange=xrange, nolost=nolost, ref=ref)
                if ref in [24, 25]: ticket['intensity'] = beam.get_column(ref, nolost=nolost).sum()

                # TODO: check congruence between tickets
                if not ticket_to_add is None:
                    last_ticket = copy.deepcopy(ticket)

                    if ticket['histogram'].size == ticket_to_add['histogram'].size:
                        ticket['histogram'] += ticket_to_add['histogram']
                        ticket['histogram_path'] += ticket_to_add['histogram_path']

                        ticket['intensity'] += ticket_to_add['intensity']
                        ticket['nrays'] += ticket_to_add['nrays']
                        ticket['good_rays'] += ticket_to_add['good_rays']

                ticket['fwhm'], ticket['fwhm_quote'], ticket['fwhm_coordinates'] = get_fwhm(ticket['histogram'], ticket['bin_center'], ret0=None)
                ticket['sigma']    = get_sigma(ticket['histogram'], ticket['bin_center'], ret0=numpy.nan)
                ticket['centroid'] = get_average(ticket['histogram'], ticket['bin_center'], ret0=numpy.nan)

                factor = ShadowPlot.get_factor(col)

                if ref != 0 and not ytitle is None:  ytitle = ytitle + ' weighted by ' + ShadowPlot.get_shadow_label(ref)

                histogram = ticket['histogram_path']
                bins      = ticket['bin_path']*factor

                if histogram.size > 1: self.plot_canvas.addCurve(bins, histogram, title, symbol='', color='blue', replace=True) #'+', '^', ','
                else:                  self.plot_canvas.clear()

                if not xtitle is None: self.plot_canvas.setGraphXLabel(xtitle)
                if not ytitle is None: self.plot_canvas.setGraphYLabel(ytitle)
                if not title is None: self.plot_canvas.setGraphTitle(title)
                self.plot_canvas.setInteractiveMode(mode='zoom')

                if ticket['fwhm'] is None: ticket['fwhm'] = 0.0
                if not ticket_to_add is None:
                    if last_ticket['fwhm'] is None: last_ticket['fwhm'] = 0.0

                n_patches = len(self.plot_canvas._backend.ax.patches)
                if (n_patches > 0): self.plot_canvas._backend.ax.patches[n_patches-1].remove()

                if not ticket['fwhm'] == 0.0:
                    x_fwhm_i, x_fwhm_f = ticket['fwhm_coordinates']
                    x_fwhm_i, x_fwhm_f = x_fwhm_i*factor, x_fwhm_f*factor
                    y_fwhm             = ticket['fwhm_quote']

                    self.plot_canvas._backend.ax.add_patch(FancyArrowPatch([x_fwhm_i, y_fwhm],
                                                                           [x_fwhm_f, y_fwhm],
                                                                           arrowstyle=ArrowStyle.CurveAB(head_width=2, head_length=4),
                                                                           color='b',
                                                                           linewidth=1.5))

                if histogram.size > 1:
                    if min(histogram) < 0: self.plot_canvas.setGraphYLimits(min(histogram), max(histogram))
                    else: self.plot_canvas.setGraphYLimits(0, max(histogram))

                self.plot_canvas.replot()

                self.info_box.intensity.setText("{:4.3f}".format(ticket['intensity']))
                self.info_box.set_flux(flux)
                self.info_box.total_rays.setText(str(ticket['nrays']))
                self.info_box.total_good_rays.setText(str(ticket['good_rays']))
                self.info_box.total_lost_rays.setText(str(ticket['nrays']-ticket['good_rays']))
                self.info_box.fwhm_h.setText("{:5.4f}".format(ticket['fwhm']*factor))
                self.info_box.label_h.setText("FWHM " + xum)
                self.info_box.sigma_h.setText("{:5.4f}".format(ticket['sigma']*factor))
                self.info_box.label_s_h.setText("\u03c3 " + xum)
                self.info_box.centroid_h.setText("{:5.4f}".format(ticket['centroid']*factor))
                self.info_box.label_c_h.setText("centroid " + xum)

                if not ticket_to_add is None:
                    return ticket, last_ticket
                else:
                    return ticket, None

            def clear(self):
                self.plot_canvas.clear()
                self.info_box.clear()

        class DetailedPlotWidget(QWidget, OWComponent):
            def __init__(self, x_scale_factor = 1.0, y_scale_factor = 1.0):
                super(ShadowPlot.DetailedPlotWidget, self).__init__()

                self.x_scale_factor = x_scale_factor
                self.y_scale_factor = y_scale_factor

                self.plot_canvas = gui.imageWiew(parent=self)

                self.plot_canvas.setColormap({"name":"temperature", "normalization":"linear", "autoscale":True, "vmin":0, "vmax":0, "colors":256})
                self.plot_canvas.setMinimumWidth(int(590 * x_scale_factor))
                self.plot_canvas.setMaximumWidth(int(590 * y_scale_factor))

                self.info_box = ShadowPlot.InfoBoxWidget(x_scale_factor, y_scale_factor)

                layout = QGridLayout()

                layout.addWidget(self.info_box,    0, 1, 1, 1)
                layout.addWidget(self.plot_canvas, 0, 0, 1, 1)

                layout.setColumnMinimumWidth(0, int(600 * x_scale_factor))
                layout.setColumnMinimumWidth(1, int(230 * x_scale_factor))

                self.setLayout(layout)

            def plot_xy(self, beam, var_x, var_y, title, xtitle, ytitle, xrange=None, yrange=None, nolost=1, nbins=100, nbins_h=None, nbins_v=None, xum="", yum="", ref=23, is_footprint=False, ticket_to_add=None, flux=None, flip_h=False, flip_v=False):
                rcParams['axes.formatter.useoffset']='False'

                if nbins_h is None: nbins_h = nbins
                if nbins_v is None: nbins_v = nbins

                ticket = beam.histo2(var_x, var_y, nbins=nbins, nbins_h=nbins_h, nbins_v=nbins_v, xrange=xrange, yrange=yrange, nolost=nolost, ref=ref)
                if ref in [24, 25]: ticket['intensity'] = beam.get_column(ref, nolost=nolost).sum()

                # TODO: check congruence between tickets
                if not ticket_to_add is None:
                    last_ticket = copy.deepcopy(ticket)

                    if ticket['histogram'].shape == ticket_to_add['histogram'].shape:
                        ticket['histogram'] += ticket_to_add['histogram']
                        ticket['histogram_h'] += ticket_to_add['histogram_h']
                        ticket['histogram_v'] += ticket_to_add['histogram_v']

                        ticket['intensity'] += ticket_to_add['intensity']
                        ticket['nrays'] += ticket_to_add['nrays']
                        ticket['good_rays'] += ticket_to_add['good_rays']

                ticket['fwhm_h'], ticket['fwhm_quote_h'], ticket['fwhm_coordinates_h'] = get_fwhm(ticket['histogram_h'], ticket['bin_h_center'], ret0=None)
                ticket['fwhm_v'], ticket['fwhm_quote_v'], ticket['fwhm_coordinates_v'] = get_fwhm(ticket['histogram_v'], ticket['bin_v_center'], ret0=None)
                ticket['sigma_h']    = get_sigma(ticket['histogram_h'], ticket['bin_h_center'], ret0=numpy.nan)
                ticket['sigma_v']    = get_sigma(ticket['histogram_v'], ticket['bin_v_center'], ret0=numpy.nan)
                ticket['centroid_h'] = get_average(ticket['histogram_h'], ticket['bin_h_center'], ret0=numpy.nan)
                ticket['centroid_v'] = get_average(ticket['histogram_v'], ticket['bin_v_center'], ret0=numpy.nan)

                factor1 = 1.0 if is_footprint else ShadowPlot.get_factor(var_x)
                factor2 = 1.0 if is_footprint else ShadowPlot.get_factor(var_y)

                xx = ticket['bin_h_edges']
                yy = ticket['bin_v_edges']

                if ticket['histogram'].size > 1:
                    xmin, xmax = xx.min(), xx.max()
                    ymin, ymax = yy.min(), yy.max()

                    origin = (xmin*factor1, ymin*factor2)
                    scale = (abs((xmax-xmin)/nbins_h)*factor1, abs((ymax-ymin)/nbins_v)*factor2)

                    self.plot_canvas.setColormap({"name":QSettings().value("output/shadow-default-colormap", "temperature", str),
                                                  "normalization":"linear",
                                                  "autoscale":True,
                                                  "vmin":0,
                                                  "vmax":0,
                                                  "colors":256})

                    # PyMCA inverts axis!!!! histogram must be calculated reversed

                    data_2D = ticket['histogram'].T
                    if flip_h: data_2D = numpy.flip(data_2D, axis=0)
                    if flip_v: data_2D = numpy.flip(data_2D, axis=1)

                    self.plot_canvas.setImage(data_2D, origin=origin, scale=scale)
                else:
                    self.plot_canvas.clear()

                if xtitle is None: xtitle=ShadowPlot.get_shadow_label(var_x)
                if ytitle is None: ytitle=ShadowPlot.get_shadow_label(var_y)

                self.plot_canvas.setGraphXLabel(xtitle)
                self.plot_canvas.setGraphYLabel(ytitle)
                self.plot_canvas.setGraphTitle(title)

                self.plot_canvas._histoHPlot.setGraphYLabel('A.U.')

                self.plot_canvas._histoHPlot._backend.ax.xaxis.get_label().set_color('white')
                self.plot_canvas._histoHPlot._backend.ax.xaxis.get_label().set_fontsize(1)
                for label in self.plot_canvas._histoHPlot._backend.ax.xaxis.get_ticklabels():
                    label.set_color('white')
                    label.set_fontsize(1)

                self.plot_canvas._histoVPlot.setGraphXLabel('A.U.')

                self.plot_canvas._histoVPlot._backend.ax.yaxis.get_label().set_color('white')
                self.plot_canvas._histoVPlot._backend.ax.yaxis.get_label().set_fontsize(1)
                for label in self.plot_canvas._histoVPlot._backend.ax.yaxis.get_ticklabels():
                    label.set_color('white')
                    label.set_fontsize(1)

                if ticket['fwhm_h'] is None: ticket['fwhm_h'] = 0.0
                if ticket['fwhm_v'] is None: ticket['fwhm_v'] = 0.0
                if not ticket_to_add is None:
                    if last_ticket['fwhm_h'] is None: last_ticket['fwhm_h'] = 0.0
                    if last_ticket['fwhm_v'] is None: last_ticket['fwhm_v'] = 0.0

                n_patches = len(self.plot_canvas._histoHPlot._backend.ax.patches)
                if (n_patches > 0): self.plot_canvas._histoHPlot._backend.ax.patches[n_patches-1].remove()

                if not ticket['fwhm_h'] == 0.0:
                    x_fwhm_i, x_fwhm_f = ticket['fwhm_coordinates_h']
                    x_fwhm_i, x_fwhm_f = x_fwhm_i*factor1, x_fwhm_f*factor1
                    y_fwhm = ticket['fwhm_quote_h']

                    self.plot_canvas._histoHPlot._backend.ax.add_patch(FancyArrowPatch([x_fwhm_i, y_fwhm],
                                                                         [x_fwhm_f, y_fwhm],
                                                                         arrowstyle=ArrowStyle.CurveAB(head_width=2, head_length=4),
                                                                         color='b',
                                                                         linewidth=1.5))

                n_patches = len(self.plot_canvas._histoVPlot._backend.ax.patches)
                if (n_patches > 0): self.plot_canvas._histoVPlot._backend.ax.patches[n_patches-1].remove()

                if not ticket['fwhm_v'] == 0.0:
                    y_fwhm_i, y_fwhm_f = ticket['fwhm_coordinates_v']
                    y_fwhm_i, y_fwhm_f = y_fwhm_i*factor2, y_fwhm_f*factor2
                    x_fwhm = ticket['fwhm_quote_v']

                    self.plot_canvas._histoVPlot._backend.ax.add_patch(FancyArrowPatch([x_fwhm, y_fwhm_i],
                                                                         [x_fwhm, y_fwhm_f],
                                                                         arrowstyle=ArrowStyle.CurveAB(head_width=2, head_length=4),
                                                                         color='r',
                                                                         linewidth=1.5))

                self.plot_canvas._histoHPlot.replot()
                self.plot_canvas._histoVPlot.replot()
                self.plot_canvas.replot()

                self.info_box.intensity.setText("{:4.3f}".format(ticket['intensity']))
                self.info_box.set_flux(flux)
                self.info_box.total_rays.setText(str(ticket['nrays']))
                self.info_box.total_good_rays.setText(str(ticket['good_rays']))
                self.info_box.total_lost_rays.setText(str(ticket['nrays']-ticket['good_rays']))
                self.info_box.fwhm_h.setText("{:5.4f}".format(ticket['fwhm_h'] * factor1))
                self.info_box.fwhm_v.setText("{:5.4f}".format(ticket['fwhm_v'] * factor2))
                self.info_box.label_h.setText("FWHM " + xum)
                self.info_box.label_v.setText("FWHM " + yum)
                self.info_box.sigma_h.setText("{:5.4f}".format(ticket['sigma_h'] * factor1))
                self.info_box.sigma_v.setText("{:5.4f}".format(ticket['sigma_v'] * factor2))
                self.info_box.label_s_h.setText("\u03c3 " + xum)
                self.info_box.label_s_v.setText("\u03c3 " + yum)
                self.info_box.centroid_h.setText("{:5.4f}".format(ticket['centroid_h'] * factor1))
                self.info_box.centroid_v.setText("{:5.4f}".format(ticket['centroid_v'] * factor2))
                self.info_box.label_c_h.setText("centroid " + xum)
                self.info_box.label_c_v.setText("centroid " + yum)

                if not ticket_to_add is None:
                    return ticket, last_ticket
                else:
                    return ticket, None

            def clear(self):
                self.plot_canvas.clear()

                self.plot_canvas._histoHPlot.clear()
                self.plot_canvas._histoVPlot.clear()

                self.plot_canvas._histoHPlot._backend.ax.xaxis.get_label().set_color('white')
                self.plot_canvas._histoHPlot._backend.ax.xaxis.get_label().set_fontsize(1)
                for label in self.plot_canvas._histoHPlot._backend.ax.xaxis.get_ticklabels():
                    label.set_color('white')
                    label.set_fontsize(1)

                self.plot_canvas._histoVPlot._backend.ax.yaxis.get_label().set_color('white')
                self.plot_canvas._histoVPlot._backend.ax.yaxis.get_label().set_fontsize(1)
                for label in self.plot_canvas._histoVPlot._backend.ax.yaxis.get_ticklabels():
                    label.set_color('white')
                    label.set_fontsize(1)

                self.plot_canvas._histoHPlot.setGraphYLabel('A.U.')
                self.plot_canvas._histoVPlot.setGraphXLabel('A.U.')

                self.plot_canvas._histoHPlot.replot()
                self.plot_canvas._histoVPlot.replot()

                self.info_box.clear()

        #########################################################################################

        @classmethod
        def plotxy_preview(cls, plot_window, beam, var_x, var_y, nolost=0, title='PLOTXY', xtitle=None, ytitle=None, is_footprint=False):
            rcParams['axes.formatter.useoffset']='False'

            col1 = beam.get_column(var_x, nolost=nolost)
            col2 = beam.get_column(var_y, nolost=nolost)

            factor1 = 1.0 if is_footprint else ShadowPlot.get_factor(var_x)
            factor2 = 1.0 if is_footprint else ShadowPlot.get_factor(var_y)

            if xtitle is None: xtitle=ShadowPlot.get_shadow_label(var_x)
            if ytitle is None: ytitle=ShadowPlot.get_shadow_label(var_y)

            plot_window.addCurve(col1*factor1, col2*factor2, title, symbol='.', color='blue', replace=True) #'+', '^', ','

            if not xtitle is None: plot_window.setGraphXLabel(xtitle)
            if not ytitle is None: plot_window.setGraphYLabel(ytitle)
            if not title is None: plot_window.setGraphTitle(title)
            plot_window.setInteractiveMode(mode='zoom')

        @classmethod
        def plot_histo_preview(cls, plot_window, beam, col, nolost, ref, title, xtitle, ytitle, conv=1.0):
            rcParams['axes.formatter.useoffset']='False'

            factor = ShadowPlot.get_factor(col)

            ticket = beam.histo1(col, nbins=100, xrange=None, nolost=nolost, ref=ref)

            if ref != 0 and not ytitle is None:  ytitle = ytitle + ' weighted by ' + ShadowPlot.get_shadow_label(ref)

            histogram = ticket['histogram_path']
            bins = ticket['bin_path']*factor

            plot_window.addCurve(bins, histogram, title, symbol='', color='blue', replace=True) #'+', '^', ','
            if not xtitle is None: plot_window.setGraphXLabel(xtitle)
            if not ytitle is None: plot_window.setGraphYLabel(ytitle)
            if not title is None: plot_window.setGraphTitle(title)
            plot_window.setInteractiveMode(mode='zoom')
            if min(histogram) < 0:
                plot_window.setGraphYLimits(min(histogram), max(histogram))
            else:
                plot_window.setGraphYLimits(0, max(histogram))
            plot_window.replot()

        @classmethod
        def get_factor(cls, var):
            if ShadowPlot.get_conversion_active() and var in [1, 2, 3, 4, 5, 6]: return 1e6 # m to micron
            else:                                                                return 1.0

        @classmethod
        def get_shadow_label(cls, var):
            return S4Beam.column_names_formatted()[var-1]
            # return (stp.getLabel(var-1))[0]


        #########################################################################################
        #
        # SAVE/LOAD FILES in HDF5 Format
        #
        #########################################################################################


        class PlotXYHdf5File(h5py.File):
            def __init__(self, file_name, mode="w"):
                try:
                    super(ShadowPlot.PlotXYHdf5File, self).__init__(name=file_name, mode=mode)
                except OSError as e:
                    if "already open" in str(e) and mode=="w":
                        super(ShadowPlot.PlotXYHdf5File, self).__init__(name=file_name, mode="a")
                        self.close()
                        super(ShadowPlot.PlotXYHdf5File, self).__init__(name=file_name, mode="w")

                if mode != "r":
                    self.coordinates = self.create_group("coordinates")
                    self.plots = self.create_group("xy_plots")
                    self.additional_data = self.create_group("additional_data")
                    self.last_plot = self.plots.create_group("last_plot")

                    self.has_last_plot = False
                    self.has_coordinate = False

                    self.attrs["default"]          = "coordinates/X"
                    self.attrs["file_name"]        = file_name
                    self.attrs["file_time"]        = time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime())
                    self.attrs["creator"]          = "PlotXYHdf5File.__init__"
                    self.attrs["code"]             = "ShadowOui"
                    self.attrs["HDF5_Version"]     = h5py.version.hdf5_version
                    self.attrs["h5py_version"]     = h5py.version.version

            def get_last_plot(self, dataset_name="intensity"):
                return self.get_plot_xy(dataset_name=dataset_name)

            def get_coordinates(self):
                bin_h_center = self["coordinates/X"][()]
                h_label = self["coordinates"].attrs["x_label"]
                bin_v_center = self["coordinates/Y"][()]
                v_label = self["coordinates"].attrs["y_label"]

                return bin_h_center, bin_v_center, h_label, v_label

            def get_plot_xy(self, plot_name="last_plot", dataset_name="intensity"):
                histogram = self["/xy_plots/" + plot_name + "/" + dataset_name][()]
                histogram_h = self["/xy_plots/" + plot_name + "/histogram_h"][()]
                histogram_v = self["/xy_plots/" + plot_name + "/histogram_v"][()]

                return histogram, histogram_h, histogram_v, self["/xy_plots/" + plot_name].attrs

            def write_coordinates(self, ticket):
                if not self.has_coordinate:
                    self.x = self.coordinates.create_dataset("X", data=ticket["bin_h_center"])
                    self.y = self.coordinates.create_dataset("Y", data=ticket["bin_v_center"])
                    self.has_coordinate = True
                else:
                    self.x[...] = ticket["bin_h_center"]
                    self.y[...] = ticket["bin_v_center"]

                try:
                    self.coordinates.attrs["x_label"] = ShadowPlot.get_shadow_label(ticket["col_h"])
                    self.coordinates.attrs["y_label"] = ShadowPlot.get_shadow_label(ticket["col_v"])
                except:
                    self.coordinates.attrs["x_label"] = ticket["h_label"]
                    self.coordinates.attrs["y_label"] = ticket["v_label"]


            def add_plot_xy(self, ticket, plot_name="last_plot", dataset_name="intensity", attributes={}):
                    if plot_name is None or plot_name.strip() == "" or plot_name.strip() == "last_plot":
                        if not self.has_last_plot:
                            self.lp_histogram   = self.last_plot.create_dataset(dataset_name, data=ticket['histogram'])
                            self.lp_histogram_h = self.last_plot.create_dataset("histogram_h", data=ticket['histogram_h'])
                            self.lp_histogram_v = self.last_plot.create_dataset("histogram_v", data=ticket['histogram_v'])
                            self.has_last_plot = True
                        else:
                            if self.lp_histogram.name != "/xy_plots/last_plot/" + dataset_name:
                                self.last_plot.move(self.lp_histogram.name, "/xy_plots/last_plot/" + dataset_name)

                            self.lp_histogram[...]   = ticket['histogram']
                            self.lp_histogram_h[...] = ticket['histogram_h']
                            self.lp_histogram_v[...] = ticket['histogram_v']

                        self.last_plot.attrs["intensity"] = ticket["intensity"]
                        self.last_plot.attrs["total_rays"] = ticket["nrays"]
                        self.last_plot.attrs["good_rays"] = ticket["good_rays"]
                        self.last_plot.attrs["lost_rays"] = ticket["nrays"]-ticket["good_rays"]

                        if not attributes is None:
                            for key in attributes.keys():
                                self.last_plot.attrs[key] = attributes[key]

                    else:
                        plot = self.plots.create_group(plot_name)

                        plot.create_dataset(dataset_name, data=ticket['histogram'])
                        plot.create_dataset("histogram_h", data=ticket['histogram_h'])
                        plot.create_dataset("histogram_v", data=ticket['histogram_v'])
                        plot.attrs["intensity"] = ticket["intensity"]
                        plot.attrs["total_rays"] = ticket["nrays"]
                        plot.attrs["good_rays"] = ticket["good_rays"]
                        plot.attrs["lost_rays"] = ticket["nrays"]-ticket["good_rays"]

                        if not attributes is None:
                            for key in attributes.keys():
                                plot.attrs[key] = attributes[key]

            def add_attribute(self, attribute_name, attribute_value, dataset_name=None):
                if not dataset_name is None:
                    self[dataset_name].attrs[attribute_name] = attribute_value
                else:
                    self.attrs[attribute_name] = attribute_value

            def get_attribute(self,  attribute_name, dataset_name=None):
                if not dataset_name is None:
                    return self[dataset_name].attrs[attribute_name]
                else:
                    return self.attrs[attribute_name]

        class HistogramHdf5File(h5py.File):
            def __init__(self, file_name, mode="w"):
                super(ShadowPlot.HistogramHdf5File, self).__init__(name=file_name, mode=mode)

                self.coordinates = self.create_group("coordinates")
                self.plots = self.create_group("histogram_plots")
                self.last_plot = self.plots.create_group("last_plot")
                self.has_last_plot = False
                self.has_coordinate = False

                self.attrs["default"]          = "coordinates/X"
                self.attrs["file_name"]        = file_name
                self.attrs["file_time"]        = time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime())
                self.attrs["creator"]          = "PlotXYHdf5File.__init__"
                self.attrs["code"]             = "ShadowOui"
                self.attrs["HDF5_Version"]     = h5py.version.hdf5_version
                self.attrs["h5py_version"]     = h5py.version.version

            def write_coordinates(self, ticket):
                if not self.has_coordinate:
                    self.x = self.coordinates.create_dataset("X", data=ticket["bin_center"])
                    self.has_coordinate = True
                else:
                    self.x[...] = ticket["bin_center"]

                self.coordinates.attrs["x_label"] = ShadowPlot.get_shadow_label(ticket["col"])

            def add_histogram(self, ticket, plot_name="last_plot", dataset_name="intensity", attributes={}):
                    if plot_name is None or plot_name.strip() == "" or plot_name.strip() == "last_plot":
                        if not self.has_last_plot:
                            self.lp_histogram  = self.last_plot.create_dataset(dataset_name, data=ticket['histogram'])
                            self.has_last_plot = True
                        else:
                            if self.lp_histogram.name != "/histogram_plots/last_plot/" + dataset_name:
                                self.last_plot.move(self.lp_histogram.name, "/histogram_plots/last_plot/" + dataset_name)

                            self.lp_histogram[...] = ticket['histogram']

                        self.last_plot.attrs["intensity"] = ticket["intensity"]
                        self.last_plot.attrs["total_rays"] = ticket["nrays"]
                        self.last_plot.attrs["good_rays"] = ticket["good_rays"]
                        self.last_plot.attrs["lost_rays"] = ticket["nrays"]-ticket["good_rays"]

                        if not attributes is None:
                            for key in attributes.keys():
                                self.last_plot.attrs[key] = attributes[key]
                    else:
                        plot = self.plots.create_group(plot_name)

                        plot.create_dataset(dataset_name, data=ticket['histogram'])
                        plot.attrs["intensity"]  = ticket["intensity"]
                        plot.attrs["total_rays"] = ticket["nrays"]
                        plot.attrs["good_rays"]  = ticket["good_rays"]
                        plot.attrs["lost_rays"]  = ticket["nrays"]-ticket["good_rays"]

                        if not attributes is None:
                            for key in attributes.keys():
                                plot.attrs[key] = attributes[key]
except:
    pass

class ShadowPreProcessor:

    @classmethod
    def read_surface_error_file(cls, filename):
        file = open(congruence.checkFile(filename), "r")

        rows = file.readlines()

        dimensions = rows[0].split()
        n_x = int(dimensions[0])
        n_y = int(dimensions[1])

        if n_x > 500:
            raise Exception("Malformed file: maximum allowed point in X direction is 500")

        x_coords = numpy.zeros(0)
        y_coords = numpy.zeros(0)
        z_values = numpy.zeros((n_x, n_y))


        index = 1
        dim_y_row = len(rows[index].split())
        is_ycoord = True
        first_x_row_index = 0

        while(is_ycoord):
            y_values = rows[index].split()

            if len(y_values) == dim_y_row:
                for y_value in y_values:
                    y_coords = numpy.append(y_coords, float(y_value))
            else:
                first_x_row_index = index
                is_ycoord = False

            index +=1

        first_x_row = rows[first_x_row_index].split()

        if len(first_x_row) == 2:
            x_index = 0
            z_index = 0

            for index in range(first_x_row_index, len(rows)):
                if z_index == 0:
                    values = rows[index].split()
                    x_coords = numpy.append(x_coords, float(values[0]))
                    z_value = float(values[1])
                else:
                    z_value = float(rows[index])

                z_values[x_index, z_index] = z_value
                z_index += 1

                if z_index == n_y:
                    x_index += 1
                    z_index = 0
        else:
            x_rows = []

            for index in range(2, len(rows)):

                x_row = rows[index].split("\t")

                if len(x_row) != 1 + n_y:
                    x_row = rows[index].split()

                if len(x_row) != 1 + n_y:
                    raise Exception("Malformed file: check format")

                x_rows.append(x_row)

            for x_index in range(0, len(x_rows)):
                x_coords = numpy.append(x_coords, float(x_rows[x_index][0]))

                for z_index in range(0, len(x_rows[x_index]) - 1):
                    z_value = float(x_rows[x_index][z_index + 1])

                    z_values[x_index, z_index] = z_value

        return x_coords, y_coords, z_values

class ShadowMath:

    @classmethod
    def gaussian_fit(cls, data_x, data_y):
        x = asarray(data_x)
        y = asarray(data_y)
        y_norm = y/sum(y)

        mean = sum(x*y_norm)
        sigma = numpy.sqrt(sum(y_norm*(x-mean)**2)/len(x))
        amplitude = max(y)

        parameters, covariance_matrix = optimize.curve_fit(ShadowMath.gaussian_function, x, y, p0 = [amplitude, mean, sigma])
        parameters.resize(4)
        parameters[3] = 2.355*parameters[2]# FWHM

        return parameters, covariance_matrix

    @classmethod
    def gaussian_function(cls, x, A, x0, sigma):
        return A*numpy.exp(-(x-x0)**2/(2*sigma**2))

    @classmethod
    def pseudovoigt_fit(cls, data_x, data_y):
        x = asarray(data_x)
        y = asarray(data_y)
        y_norm = y/sum(y)

        amplitude = max(data_y)
        mean = sum(x*y_norm)
        fwhm = numpy.sqrt(sum(y_norm*(x-mean)**2)/len(x))*2.355
        mixing = 0.1

        parameters, covariance_matrix = optimize.curve_fit(ShadowMath.pseudovoigt_function,
                                                           x, y,
                                                           p0 = [amplitude, mean, fwhm, mixing],
                                                           bounds = ([0.1*amplitude, 0.9*mean,  0.1*fwhm, 1e-3],
                                                                     [10.0*amplitude, 1.1*mean, 10.0*fwhm, 1.0]))

        return parameters, covariance_matrix

    @classmethod
    def pseudovoigt_function(cls, x, A, x0, fwhm, mixing):
        sigma = fwhm/2.355
        gamma = fwhm/2

        return A*(mixing*numpy.exp(-(x-x0)**2/(2*sigma**2)) + (1-mixing)*((gamma**2)/((x-x0)**2 + gamma**2)))

    @classmethod
    def caglioti_broadening_fit(cls, data_x, data_y):
        x = asarray(data_x)
        y = asarray(data_y)

        parameters, covariance_matrix = optimize.curve_fit(ShadowMath.caglioti_broadening_function,
                                                           x, y,
                                                           p0=[0.0001, 0.0001, 0.0001],
                                                           bounds = ([ -1.0, -1.0, -1.0],
                                                                     [ 1.0,  1.0,  1.0]))

        return parameters, covariance_matrix

    @classmethod
    def caglioti_broadening_function(cls, x, U, V, W):
        return numpy.sqrt(W + V * (numpy.tan(x*numpy.pi/360)) + U * (numpy.tan(x*numpy.pi/360))**2)

    @classmethod
    def caglioti_shape_fit(cls, data_x, data_y):
        x = asarray(data_x)
        y = asarray(data_y)

        parameters, covariance_matrix = optimize.curve_fit(ShadowMath.caglioti_shape_function,
                                                           x, y,
                                                           p0=[0.1, 0.0, 0.0],
                                                           bounds = ([ 0.0, -1.0, -1.0],
                                                                     [ 1.0,  1.0,  1.0]))

        return parameters, covariance_matrix

    @classmethod
    def caglioti_shape_function(cls, x, a, b, c):
        return a + b*(x*numpy.pi/360) + c*(x*numpy.pi/360)**2


    @classmethod
    def vectorial_product(cls, vector1, vector2):
        result = [0.0, 0.0, 0.0]

        result[0] = vector1[1]*vector2[2] - vector1[2]*vector2[1]
        result[1] = -(vector1[0]*vector2[2] - vector1[2]*vector2[0])
        result[2] = vector1[0]*vector2[1] - vector1[1]*vector2[0]

        return result

    @classmethod
    def scalar_product(cls, vector1, vector2):
        return vector1[0]*vector2[0] + vector1[1]*vector2[1] + vector1[2]*vector2[2]

    @classmethod
    def vector_modulus(cls, vector):
        return numpy.sqrt(cls.scalar_product(vector, vector))

    @classmethod
    def vector_multiply(cls, vector, constant):
        result = [0.0, 0.0, 0.0]

        result[0] = vector[0] * constant
        result[1] = vector[1] * constant
        result[2] = vector[2] * constant

        return result

    @classmethod
    def vector_divide(cls, vector, constant):
        result = [0.0, 0.0, 0.0]

        result[0] = vector[0] / constant
        result[1] = vector[1] / constant
        result[2] = vector[2] / constant

        return result

    @classmethod
    def vector_normalize(cls, vector):
        return cls.vector_divide(vector, cls.vector_modulus(vector))

    @classmethod
    def vector_sum(cls, vector1, vector2):
        result = [0.0, 0.0, 0.0]

        result[0] = vector1[0] + vector2[0]
        result[1] = vector1[1] + vector2[1]
        result[2] = vector1[2] + vector2[2]

        return result

    @classmethod
    def vector_difference(cls, vector1, vector2):
        result = [0.0, 0.0, 0.0]

        result[0] = vector1[0] - vector2[0]
        result[1] = vector1[1] - vector2[1]
        result[2] = vector1[2] - vector2[2]

        return result

    ##########################################################################
    # Rodrigues Formula:
    #
    # rotated = vector * cos(rotation_angle) + 
    #           (rotation_axis x vector) * sin(rotation_angle) + 
    #           rotation_axis*(rotation_axis . vector)(1 - cos(rotation_angle))
    #
    # rotation_angle in radians
    #
    ##########################################################################
    @classmethod
    def vector_rotate(cls, rotation_axis, rotation_angle, vector):

        result_temp_1 = ShadowMath.vector_multiply(vector, numpy.cos(rotation_angle))
        result_temp_2 = ShadowMath.vector_multiply(ShadowMath.vectorial_product(rotation_axis, vector),
                                                   numpy.sin(rotation_angle))
        result_temp_3 = ShadowMath.vector_multiply(ShadowMath.vector_multiply(rotation_axis, ShadowMath.scalar_product(rotation_axis, vector)), (1 - numpy.cos(rotation_angle)))

        result = ShadowMath.vector_sum(result_temp_1,
                                       ShadowMath.vector_sum(result_temp_2, result_temp_3))
        return result

    @classmethod
    def point_distance(cls, point1, point2):
        return cls.vector_modulus(cls.vector_difference(point1, point2))

class ShadowPhysics:

    A2EV = (codata.h*codata.c/codata.e)*1e+10
    K2EV = 2*numpy.pi/(codata.h*codata.c/codata.e*1e+2)

    @classmethod
    def getWavelengthFromShadowK(cls, k_mod): # in cm
        return (2*numpy.pi/k_mod)*1e+8 # in Angstrom

    @classmethod
    def getShadowKFromWavelength(cls, wavelength): # in A
        return (2*numpy.pi/wavelength)*1e+8 # in cm

    @classmethod
    def getWavelengthFromEnergy(cls, energy): #in eV
        return cls.A2EV/energy # in Angstrom

    @classmethod
    def getEnergyFromWavelength(cls, wavelength): # in Angstrom
        return cls.A2EV/wavelength # in eV

    @classmethod
    def getEnergyFromShadowK(cls, k_mod): # in cm
        return  k_mod/cls.K2EV # in eV

    @classmethod
    def getShadowKFromEnergy(cls, energy): # in A
        return cls.K2EV*energy # in cm

    @classmethod
    def calculateBraggAngle(cls, wavelength, h, k, l, a):

        # lambda = 2 pi / |k| = 2 d sen(th)
        #
        # sen(th) = lambda / (2  d)
        #
        # d = a / sqrt(h\u00b2 + k^2 + l^2)
        #
        # sen(th) = (sqrt(h^2 + k^2 + l^2) * lambda)/(2 a)

        theta_bragg = -1

        argument = wavelength*numpy.sqrt(h**2+k**2+l**2)/(2*a)
        if argument <= 1:
            result = numpy.arcsin(argument)
            if result > 0: theta_bragg = result

        return theta_bragg

    @classmethod
    def checkCompoundName(cls, compound_name):
        if compound_name is None: raise Exception("Compound Name is Empty")
        if str(compound_name.strip()) == "": raise Exception("Compound Name is Empty")

        compound_name = compound_name.strip()

        try:
            ml.CompoundParser(compound_name)
            return compound_name
        except:
            raise Exception("Compound Name is not correct")

    @classmethod
    def getMaterialDensity(cls, material_name):
        if material_name is None: return 0.0
        if str(material_name.strip()) == "": return 0.0

        try:
            compoundData = ml.CompoundParser(material_name)

            n_elements = compoundData["nElements"]
            if n_elements == 1:
                return ml.ElementDensity(compoundData["Elements"][0])
            else:
                density = 0.0
                mass_fractions = compoundData["massFractions"]
                elements = compoundData["Elements"]
                for i in range(n_elements): density += ml.ElementDensity(elements[i]) * mass_fractions[i]
                return density
        except:
            return 0.0

    @classmethod
    def ConstatoBackgroundNoised(cls, constant_value=0, n_sigma=1.0, random_generator=random.Random()):
        sigma = numpy.sqrt(constant_value) # poisson statistic
        noise = (n_sigma*sigma)*random_generator.random()
        sign_marker = random_generator.random()

        if sign_marker > 0.5:
            return int(round(constant_value+noise, 0))
        else:
            return int(round(constant_value-noise, 0))

    @classmethod
    def Chebyshev(cls, n, x):
        if n==0: return 1
        elif n==1: return x
        else: return 2*x*cls.Chebyshev(n-1, x)-cls.Chebyshev(n-2, x)

    @classmethod
    def ChebyshevBackground(cls, coefficients=[0,0,0,0,0,0], twotheta=0):
        coefficients_set = range(0, len(coefficients))
        background = 0

        for index in coefficients_set:
            background += coefficients[index]*cls.Chebyshev(index, twotheta)

        return background

    @classmethod
    def ChebyshevBackgroundNoised(cls, coefficients=[0,0,0,0,0,0], twotheta=0.0, n_sigma=1.0, random_generator=random.Random()):
        background = cls.ChebyshevBackground(coefficients, twotheta)
        sigma = numpy.sqrt(background) # poisson statistic

        noise = (n_sigma*sigma)*random_generator.random()
        sign_marker = random_generator.random()

        if sign_marker > 0.5:
            return int(round(background+noise, 0))
        else:
            return int(round(background-noise, 0))

    @classmethod
    def ExpDecay(cls, h, x):
      return numpy.exp(-h*x)

    @classmethod
    def ExpDecayBackground(cls, coefficients=[0,0,0,0,0,0], decayparams=[0,0,0,0,0,0], twotheta=0):
        coefficients_set = range(0, len(coefficients))
        background = 0

        for index in coefficients_set:
            background += coefficients[index]*cls.ExpDecay(decayparams[index], twotheta)

        return background

    @classmethod
    def ExpDecayBackgroundNoised(cls, coefficients=[0,0,0,0,0,0], decayparams=[0,0,0,0,0,0], twotheta=0, n_sigma=1, random_generator=random.Random()):
        background = cls.ExpDecayBackground(coefficients, decayparams, twotheta)
        sigma = numpy.sqrt(background) # poisson statistic

        noise = (n_sigma*sigma)*random_generator.random()
        sign_marker = random_generator.random()

        if sign_marker > 0.5:
            return int(round(background+noise, 0))
        else:
            return int(round(background-noise, 0))


from orangewidget.widget import Input, Output
from orangecontrib.shadow4.util.shadow4_objects import ShadowData

class TriggerToolsDecorator(object):
    @classmethod
    def get_trigger_input(cls):
        return Input("Trigger", TriggerOut, default=True, auto_summary=False)

    @classmethod
    def get_trigger_output(cls):
       return Output("Trigger", TriggerIn, default=True, auto_summary=False)

    def set_trigger_parameters_for_sources(self, trigger):
        if trigger and trigger.new_object == True:
            if trigger.has_additional_parameter("seed_increment"):
                self.seed += trigger.get_additional_parameter("seed_increment")

            if trigger.has_additional_parameter("variable_name"):
                variable_name         = trigger.get_additional_parameter("variable_name").strip()
                variable_display_name = trigger.get_additional_parameter("variable_display_name").strip()
                variable_value        = trigger.get_additional_parameter("variable_value")
                variable_um           = trigger.get_additional_parameter("variable_um")

                def check_number(x):
                    try:    return float(x)
                    except: return x

                if "," in variable_name:
                    variable_names = variable_name.split(",")

                    if isinstance(variable_value, str) and "," in variable_value:
                        variable_values = variable_value.split(",")
                        for variable_name, variable_value in zip(variable_names, variable_values):
                            setattr(self, variable_name.strip(), check_number(variable_value))
                            self.check_options(variable_name)
                    else:
                        for variable_name in variable_names:
                            setattr(self, variable_name.strip(), check_number(variable_value))
                            self.check_options(variable_name)
                else:
                    setattr(self, variable_name, check_number(variable_value))
                    self.check_options(variable_name)

                scanning_data = ShadowData.ScanningData(variable_name, variable_value, variable_display_name, variable_um)
            else:
                scanning_data = None

            self.run_shadow4(scanning_data=scanning_data)

    def check_options(self, variable_name):
        pass

    def set_trigger_parameters_for_optics(self, trigger): # TODO: complete
        if ShadowCongruence.check_empty_data(self.input_data):
            if trigger and trigger.new_object == True:
                if trigger.has_additional_parameter("variable_name"):
                    variable_name         = trigger.get_additional_parameter("variable_name").strip()
                    variable_display_name = trigger.get_additional_parameter("variable_display_name").strip()
                    variable_value        = trigger.get_additional_parameter("variable_value")
                    variable_um           = trigger.get_additional_parameter("variable_um")

                    def check_number(x):
                        try:    return float(x)
                        except: return x

                    if "," in variable_name:
                        variable_names = variable_name.split(",")

                        if isinstance(variable_value, str) and "," in variable_value:
                            variable_values = variable_value.split(",")
                            for variable_name, variable_value in zip(variable_names, variable_values):
                                setattr(self, variable_name.strip(), check_number(variable_value))
                                self.check_options(variable_name)
                        else:
                            for variable_name in variable_names:
                                setattr(self, variable_name.strip(), check_number(variable_value))
                                self.check_options(variable_name)
                    else:
                        setattr(self, variable_name, check_number(variable_value))
                        self.check_options(variable_name)

                    scanning_data = ShadowData.ScanningData(variable_name, variable_value, variable_display_name, variable_um)
                else:
                    scanning_data = None

                self.run_shadow4(scanning_data=scanning_data)

class Properties(object):
    def __init__(self, props=None):
        self._props = {}
        self._origprops = {}
        self._keymap = {}
        self.othercharre = re.compile(r'(?<!\\)(\s*\=)|(?<!\\)(\s*\:)')
        self.othercharre2 = re.compile(r'(\s*\=)|(\s*\:)')
        self.bspacere = re.compile(r'\\(?!\s$)')

    def __str__(self):
        s='{'
        for key,value in self._props.items():
            s = ''.join((s,key,'=',value,', '))

        s=''.join((s[:-2],'}'))
        return s

    def __parse(self, lines):
        # Every line in the file must consist of either a comment
        # or a key-value pair. A key-value pair is a line consisting
        # of a key which is a combination of non-white space characters
        # The separator character between key-value pairs is a '=',
        # ':' or a whitespace character not including the newline.
        # If the '=' or ':' characters are found, in the line, even
        # keys containing whitespace chars are allowed.
        # A line with only a key according to the rules above is also
        # fine. In such case, the value is considered as the empty string.
        # In order to include characters '=' or ':' in a key or value,
        # they have to be properly escaped using the backslash character.

        # Some examples of valid key-value pairs:
        #
        # key     value
        # key=value
        # key:value
        # key     value1,value2,value3
        # key     value1,value2,value3 \
        #         value4, value5
        # key
        # This key= this value
        # key = value1 value2 value3

        # Any line that starts with a '#' is considered a comment
        # and skipped. Also any trailing or preceding whitespaces
        # are removed from the key/value.

        # This is a line parser. It parses the
        # contents like by line.

        lineno=0
        i = iter(lines)

        for line in i:
            lineno += 1
            line = line.strip()

            if not line: continue
            if line[0] == '#': continue

            sepidx = -1

            m = self.othercharre.search(line)
            if m:
                first, last = m.span()
                start, end = 0, first
                wspacere = re.compile(r'(?<![\\\=\:])(\s)')
            else:
                if self.othercharre2.search(line):
                    wspacere = re.compile(r'(?<![\\])(\s)')

                start, end = 0, len(line)

            m2 = wspacere.search(line, start, end)
            if m2:
                first, last = m2.span()
                sepidx = first
            elif m:
                first, last = m.span()
                sepidx = last - 1

            while line[-1] == '\\':
                nextline = i.next()
                nextline = nextline.strip()
                lineno += 1
                line = line[:-1] + nextline

            if sepidx != -1:
                key, value = line[:sepidx], line[sepidx+1:]
            else:
                key,value = line,''

            self.processPair(key, value)

    def processPair(self, key, value):
        oldkey = key
        oldvalue = value

        keyparts = self.bspacere.split(key)

        strippable = False
        lastpart = keyparts[-1]

        if lastpart.find('\\ ') != -1:
            keyparts[-1] = lastpart.replace('\\','')

        elif lastpart and lastpart[-1] == ' ':
            strippable = True

        key = ''.join(keyparts)
        if strippable:
            key = key.strip()
            oldkey = oldkey.strip()

        oldvalue = self.unescape(oldvalue)
        value = self.unescape(value)

        self._props[key] = value.strip()

        if self._keymap.__contains__(key):
            oldkey = self._keymap.get(key)
            self._origprops[oldkey] = oldvalue.strip()
        else:
            self._origprops[oldkey] = oldvalue.strip()
            self._keymap[key] = oldkey

    def escape(self, value):
        newvalue = value.replace(':','\\:')
        newvalue = newvalue.replace('=','\\=')

        return newvalue

    def unescape(self, value):
        newvalue = value.replace('\\:',':')
        newvalue = newvalue.replace('\\=','=')

        return newvalue

    def load(self, stream):
        if not hasattr(stream, 'read'):
            raise TypeError('Argument should be a file object!')
        if stream.mode != 'r':
            raise ValueError ('Stream should be opened in read-only mode!')

        try:
            lines = stream.readlines()
            self.__parse(lines)
        except IOError as e:
            raise e

    def getProperty(self, key):
        return self._props.get(key)

    def setProperty(self, key, value):
        if type(key) is str and type(value) is str:
            self.processPair(key, value)
        else:
            raise TypeError('both key and value should be strings!')

    def propertyNames(self):
        return self._props.keys()

    def list(self, out=sys.stdout):
        out.write('-- listing properties --\n')
        for key,value in self._props.items():
            out.write(''.join((key,'=',value,'\n')))

    def store(self, out, header=""):
        if out.mode[0] != 'w':
            raise ValueError('Steam should be opened in write mode!')

        try:
            out.write(''.join(('#',header,'\n')))
            tstamp = time.strftime('%a %b %d %H:%M:%S %Z %Y', time.localtime())
            out.write(''.join(('#',tstamp,'\n')))
            for prop, val in self._origprops.items():
                out.write(''.join((prop,'=',self.escape(val),'\n')))

            out.close()
        except IOError as e:
            raise e

    def getPropertyDict(self):
        return self._props

    def __getitem__(self, name):
        return self.getProperty(name)

    def __setitem__(self, name, value):
        self.setProperty(name, value)

    def __getattr__(self, name):
        try:
            return self.__dict__[name]
        except KeyError:
            if hasattr(self._props,name):
                return getattr(self._props, name)

try:
    from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
    from matplotlib.figure import Figure
    from mpl_toolkits.mplot3d import Axes3D  # necessario per caricare i plot 3D
except:
    pass

try:
    from AnyQt.QtWidgets import QApplication, QVBoxLayout
    from AnyQt.QtCore import QCoreApplication

    class MathTextLabel(QWidget):
        def __init__(self, mathText, size=None, parent=None, **kwargs):
            QWidget.__init__(self, parent, **kwargs)

            l=QVBoxLayout(self)
            l.setContentsMargins(0,0,0,0)

            r,g,b,a=self.palette().base().color().getRgbF()

            self._figure=Figure(edgecolor=(r,g,b), facecolor=(r,g,b))
            self._canvas=FigureCanvas(self._figure)
            l.addWidget(self._canvas)

            self._figure.clear()

            if not size:
                size = QCoreApplication.instance().font().pointSize()

            text=self._figure.suptitle(
                mathText,
                x=0.0,
                y=1.0,
                horizontalalignment='left',
                verticalalignment='top',
                size=size)
            self._canvas.draw()

            (x0,y0),(x1,y1)=text.get_window_extent().get_points()
            w=x1-x0; h=y1-y0

            self._figure.set_size_inches(w/80, h/80)
            self.setFixedSize(w,h)
except:
    pass


if __name__ == "__main__":

    try:
        import matplotlib
        import matplotlib.pyplot as plt
        from matplotlib import cm

    except ImportError:
        print(sys.exc_info()[1])
        pass

    #print(congruence.checkFileName("pippo.dat"))
    #print(congruence.checkFileName("Files/pippo.dat"))
    #print(congruence.checkFileName("Files/pippo.dat"))
    #print(congruence.checkFileName("/Users/labx/Desktop/pippo.dat"))

    s = " 5   8095683980.2420149       3.34799999999999994E-008"

    print(s.strip().split(" "))


    print("Bragg")
    ShadowCongruence.checkBraggFile("/Users/labx/Oasys/bragg.dat")
    print("PreRefl")
    ShadowCongruence.checkPreReflFile("/Users/labx/Oasys/reflec.dat")
    print("PreMLayer")
    ShadowCongruence.checkPreMLayerFile("/Users/labx/Oasys/mlayer.dat")

    #ShadowCongruence.checkXOPDiffractionProfileFile("/Users/labx/Oasys/mlayer.dat")

    '''
    print(ShadowPhysics.A2EV)

    print(ShadowPhysics.Chebyshev(4, 21))
    print(ShadowPhysics.Chebyshev(0, 35))

    coefficients = [5.530814e+002, 2.487256e+000, -2.004860e-001, 2.246427e-003, -1.044517e-005, 1.721576e-008]
    random_generator=random.Random()

    print(ShadowPhysics.ChebyshevBackgroundNoised(coefficients, 10, random_generator=random_generator))
    print(ShadowPhysics.ChebyshevBackgroundNoised(coefficients, 11, random_generator=random_generator))
    print(ShadowPhysics.ChebyshevBackgroundNoised(coefficients, 12, random_generator=random_generator))
    print(ShadowPhysics.ChebyshevBackgroundNoised(coefficients, 13, random_generator=random_generator))
    print(ShadowPhysics.ChebyshevBackgroundNoised(coefficients, 14, random_generator=random_generator))
    print(ShadowPhysics.ChebyshevBackgroundNoised(coefficients, 15, random_generator=random_generator))
    print(ShadowPhysics.ChebyshevBackgroundNoised(coefficients, 16, random_generator=random_generator))
    print(ShadowPhysics.ChebyshevBackgroundNoised(coefficients, 17, random_generator=random_generator))
    print(ShadowPhysics.ChebyshevBackgroundNoised(coefficients, 18, random_generator=random_generator))
    print(ShadowPhysics.ChebyshevBackgroundNoised(coefficients, 19, random_generator=random_generator))
    print(ShadowPhysics.ChebyshevBackgroundNoised(coefficients, 20, random_generator=random_generator))
    '''
    '''
    import matplotlib.pyplot as plt

    x_coords, y_coords, z_values = ShadowPreProcessor.read_surface_error_file("/Users/labx/Oasys/mirror.dat")

    fig = plt.figure()
    ax = fig.gca(projection='3d')
    X, Y = numpy.meshgrid(x_coords, y_coords)
    surf = ax.plot_surface(X, Y, z_values.T, rstride=1, cstride=1, cmap=cm.coolwarm,
            linewidth=0, antialiased=False)
    #ax.set_zlim(-1.01, 1.01)

    fig.colorbar(surf, shrink=0.5, aspect=5)

    plt.show()

    app = QApplication(sys.argv)

    widget = QWidget()
    widget.setLayout(QVBoxLayout())

    figure = Figure(figsize=(100, 100))
    figure.patch.set_facecolor('white')

    axis = figure.add_subplot(111, projection='3d')

    axis.set_xlabel("X (cm)")
    axis.set_ylabel("Y (cm)")
    axis.set_zlabel("Z (cm)")

    figure_canvas = FigureCanvasQTAgg(figure)
    widget.layout().addWidget(figure_canvas)
    figure_canvas.setFixedWidth(500)
    figure_canvas.setFixedHeight(450)

    x_coords, y_coords, z_values = ShadowPreProcessor.read_surface_error_file("/Users/labx/Oasys/mirror.dat")

    x_to_plot, y_to_plot = numpy.meshgrid(x_coords, y_coords)

    axis.plot_surface(x_to_plot, y_to_plot, z_values.T,
                      rstride=1, cstride=1, cmap=cm.autumn, linewidth=0.5, antialiased=True)

    figure_canvas.draw()
    figure_canvas.show()
    widget.show()

    app.exec()
    '''




import numpy

from orangewidget.settings import Setting
from orangewidget import gui as orangegui

from oasys2.widget import gui as oasysgui
from oasys2.widget.util import congruence
from oasys2.canvas.util.canvas_util import add_widget_parameters_to_module

from syned.storage_ring.magnetic_structures.bending_magnet import BendingMagnet

from shadow4.sources.bending_magnet.s4_bending_magnet import S4BendingMagnet
from shadow4.sources.bending_magnet.s4_bending_magnet_light_source import S4BendingMagnetLightSource

from orangecontrib.shadow4.widgets.gui.ow_synchrotron_source import OWSynchrotronSource
from orangecontrib.shadow4.widgets.gui.plots import plot_data1D

class OWBendingMagnet(OWSynchrotronSource):
    name = "Bending Magnet"
    description = "Shadow Source: Bending Magnet"
    icon = "icons/bending_magnet.png"
    priority = 3

    magnetic_field         = Setting(-1.26754)
    divergence             = Setting(69e-3)
    emin                   = Setting(1000.0)  # Photon energy scan from energy (in eV)
    emax                   = Setting(1000.1)  # Photon energy scan to energy (in eV)
    ng_e                   = Setting(100)     # Photon energy scan number of points

    plot_bm_graph = 1

    def __init__(self):
        super().__init__()

        tab_bas = oasysgui.createTabPage(self.tabs_control_area, "Bending Magnet")

        #
        box_1 = oasysgui.widgetBox(tab_bas, "Bending Magnet Parameters", addSpace=True, orientation="vertical")
        oasysgui.lineEdit(box_1, self, "emin", "Minimum Energy [eV]", tooltip="emin", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(box_1, self, "emax", "Maximum Energy [eV]", tooltip="emax", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(box_1, self, "magnetic_field", "Magnetic Field [T]", tooltip="magnetic_field", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(box_1, self, "divergence", "Horizontal divergence (arc of radius) [rads]", tooltip="divergence", labelWidth=260, valueType=float, orientation="horizontal")

        box_2 = oasysgui.widgetBox(tab_bas, "Sampling rays", addSpace=True, orientation="vertical")
        oasysgui.lineEdit(box_2, self, "number_of_rays", "Number of Rays", tooltip="Number of Rays", labelWidth=250, valueType=int, orientation="horizontal")
        oasysgui.lineEdit(box_2, self, "seed", "Seed", tooltip="Seed (0=clock)", labelWidth=250, valueType=int, orientation="horizontal")


        # bm adv settings
        tab_advanced = oasysgui.createTabPage(self.tabs_control_area, "Advanced")

        box_3 = oasysgui.widgetBox(tab_advanced, "Advanced settings", addSpace=True, orientation="vertical")
        oasysgui.lineEdit(box_3, self, "ng_e", "Spectrum number of points", tooltip="ng_e", labelWidth=250, valueType=int, orientation="horizontal")

        self.add_specific_bm_plots()

        orangegui.rubber(self.controlArea)
        orangegui.rubber(self.mainArea)

    def add_specific_bm_plots(self):
        bm_plot_tab = oasysgui.widgetBox(self.main_tabs, addToLayout=0, margin=4)

        self.main_tabs.insertTab(1, bm_plot_tab, "BM Plots")

        view_box = oasysgui.widgetBox(bm_plot_tab, "Plotting Style", addSpace=False, orientation="horizontal")
        view_box_1 = oasysgui.widgetBox(view_box, "", addSpace=False, orientation="vertical", width=350)

        self.bm_view_type_combo = orangegui.comboBox(view_box_1, self,
                                            "plot_bm_graph",
                                                          label="Plot Graphs?",
                                                          labelWidth=220,
                                                          items=["No", "Yes"],
                                                          callback=self.refresh_specific_plots,
                                                          sendSelectedValue=False,
                                                          orientation="horizontal")

        self.bm_tab = []
        self.bm_tabs = oasysgui.tabWidget(bm_plot_tab)

        current_tab = self.bm_tabs.currentIndex()

        size = len(self.bm_tab)
        indexes = range(0, size)
        for index in indexes:
            self.bm_tabs.removeTab(size-1-index)

        self.bm_tab = [
            orangegui.createTabPage(self.bm_tabs, "BM Spectrum"),
            orangegui.createTabPage(self.bm_tabs, "BM Spectral power")
        ]

        for tab in self.bm_tab:
            tab.setFixedHeight(self.IMAGE_HEIGHT)
            tab.setFixedWidth(self.IMAGE_WIDTH)

        self.bm_plot_canvas = [None, None]

        self.bm_tabs.setCurrentIndex(current_tab)

    def refresh_specific_plots(self):
        if self.plot_bm_graph == 0:
            for bm_plot_slot_index in range(6):
                current_item = self.bm_tab[bm_plot_slot_index].layout().itemAt(0)
                self.bm_tab[bm_plot_slot_index].layout().removeItem(current_item)
                plot_widget_id = oasysgui.QLabel() # TODO: is there a better way to clean this??????????????????????
                self.bm_tab[bm_plot_slot_index].layout().addWidget(plot_widget_id)
        else:
            if self.light_source is None: return

            e, f, w = self.light_source.calculate_spectrum()

            self.plot_widget_item(e, f, 0,
                                  title="BM spectrum (current = %5.1f)"%self.ring_current,
                                  xtitle="Photon energy [eV]",ytitle=r"Photons/s/0.1%bw")

            self.plot_widget_item(e, w, 1,
                                  title="BM spectrum (current = %5.1f)"%self.ring_current,
                                  xtitle="Photon energy [eV]",ytitle="Spectral power [W/eV]")

    def plot_widget_item(self,x,y,bm_plot_slot_index,title="",xtitle="",ytitle=""):
        self.bm_tab[bm_plot_slot_index].layout().removeItem(self.bm_tab[bm_plot_slot_index].layout().itemAt(0))
        plot_widget_id = plot_data1D(x.copy(),y.copy(),title=title,xtitle=xtitle,ytitle=ytitle,symbol='.')
        self.bm_tab[bm_plot_slot_index].layout().addWidget(plot_widget_id)

    def check_magnetic_structure(self):
        congruence.checkNumber(self.magnetic_field, "Magnetic Field")
        congruence.checkStrictlyPositiveNumber(self.divergence, "Divergence")

    def build_light_source(self, electron_beam, flag_emittance):
        magnetic_radius = numpy.abs(S4BendingMagnet.calculate_magnetic_radius(self.magnetic_field, electron_beam.energy()))
        length          = numpy.abs(self.divergence * magnetic_radius)

        print(">>> calculated magnetic_radius = S4BendingMagnet.calculate_magnetic_radius(%f, %f) = %f" %\
              (self.magnetic_field, electron_beam.energy(), magnetic_radius))

        print(">>> calculated BM length = divergence * magnetic_radius = %f " % length)

        bm = S4BendingMagnet(magnetic_radius,
                             self.magnetic_field,
                             length,
                             emin=self.emin,  # Photon energy scan from energy (in eV)
                             emax=self.emax,  # Photon energy scan to energy (in eV)
                             ng_e=self.ng_e,  # Photon energy scan number of points
                             flag_emittance=flag_emittance,  # when sampling rays: Use emittance (0=No, 1=Yes)
                             )

        print("\n\n***** BM info: ", bm.info())

        # S4UndulatorLightSource
        try:    name = self.getNode().title
        except: name = "Bending Magnet"

        light_source = S4BendingMagnetLightSource(name=name,
                                                  electron_beam=electron_beam,
                                                  magnetic_structure=bm,
                                                  nrays=self.number_of_rays,
                                                  seed=self.seed)

        print("\n\n***** S4BendingMagnetLightSource info: ", light_source.info())

        return light_source

    def populate_fields_from_magnetic_structure(self, magnetic_structure, electron_beam):
        if isinstance(magnetic_structure, BendingMagnet):
            self.magnetic_field = magnetic_structure.magnetic_field()
            self.divergence     = magnetic_structure.horizontal_divergence()

add_widget_parameters_to_module(__name__)



import numpy

from orangewidget.settings import Setting
from orangewidget import gui as orangegui

from oasys2.widget import gui as oasysgui
from oasys2.widget.util import congruence
from oasys2.canvas.util.canvas_util import add_widget_parameters_to_module

from syned.storage_ring.magnetic_structures.undulator import Undulator

from shadow4.sources.undulator.s4_undulator_gaussian import S4UndulatorGaussian
from shadow4.sources.undulator.s4_undulator_gaussian_light_source import S4UndulatorGaussianLightSource

from orangecontrib.shadow4.widgets.gui.ow_synchrotron_source import OWSynchrotronSource
from orangecontrib.shadow4.widgets.gui.plots import plot_data1D, plot_multi_data1D

class OWUndulatorGaussian(OWSynchrotronSource):
    name = "Undulator Gaussian"
    description = "Shadow Source: Undulator Gaussian"
    icon = "icons/ugaussian.png"
    priority = 5

    undulator_length = Setting(4.0)
    energy           = Setting(15000.0)
    delta_e          = Setting(0)

    plot_undulator_graph = Setting(1)

    # energy spread...
    harmonic_number = Setting(1)
    period_length = Setting(0.020)

    flag_autoset_flux_central_cone = Setting(0)
    flux_central_cone = Setting('1e10')
    plot_npoints = Setting(100)
    plot_Kmin = Setting(0.2)
    plot_Kmax = Setting(2.0)
    plot_max_harmonic_number = Setting(11)



    def __init__(self):
        self.__is_init = True

        super().__init__(show_energy_spread=True)

        tab_bas = oasysgui.createTabPage(self.tabs_control_area, "Undulator")

        #
        box_1 = oasysgui.widgetBox(tab_bas, "Undulator Parameters", addSpace=True, orientation="vertical")
        oasysgui.lineEdit(box_1, self, "undulator_length", "Undulator Length [m]", labelWidth=250, tooltip="undulator_length", valueType=float, orientation="horizontal")
        self.box_energy_spread_local = oasysgui.widgetBox(box_1, orientation="vertical")
        oasysgui.lineEdit(self.box_energy_spread_local, self, "period_length", "ID period [m]", labelWidth=260, tooltip="period_length", valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.box_energy_spread_local, self, "harmonic_number", "Harmonic in use [odd]", labelWidth=250, tooltip="harmonic_number", valueType=int, orientation="horizontal")


        #
        box_2 = oasysgui.widgetBox(tab_bas, "Sampling rays", addSpace=True, orientation="vertical")
        oasysgui.lineEdit(box_2, self, "energy", "Set photon energy [eV]", tooltip="energy", labelWidth=250, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(box_2, self, "delta_e", "Delta Energy [eV]", tooltip="delta_e", labelWidth=250, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(box_2, self, "number_of_rays", "Number of Rays", tooltip="number_of_rays", labelWidth=250, valueType=int, orientation="horizontal")
        oasysgui.lineEdit(box_2, self, "seed", "Seed", tooltip="seed", labelWidth=250, valueType=int, orientation="horizontal")

        #
        # advanced settings
        #
        tab_advanced = oasysgui.createTabPage(self.tabs_control_area, "Advanced")
        # arrays
        left_box_11 = oasysgui.widgetBox(tab_advanced, "Flux normalization", addSpace=False, orientation="vertical")

        orangegui.comboBox(left_box_11, self,
                           "flag_autoset_flux_central_cone",
                           label="Auto calculate central cone flux",
                           labelWidth=220,
                           items=["No", "Yes"],
                           callback=self.set_visibility,
                           sendSelectedValue=False,
                           orientation="horizontal")

        self.box_flux_central_cone = oasysgui.widgetBox(left_box_11, orientation="vertical")
        oasysgui.lineEdit(self.box_flux_central_cone, self, "flux_central_cone", "Set flux to [photons/s/0.1%bw]", tooltip="flux_central_cone",
                          labelWidth=250, valueType=str, orientation="horizontal")


        left_box_11 = oasysgui.widgetBox(tab_advanced, "Specific parameters for undulator plots", addSpace=False, orientation="vertical")
        oasysgui.lineEdit(left_box_11, self, "plot_npoints", "Number of energy points", tooltip="plot_npoints",
                          labelWidth=250, valueType=int, orientation="horizontal")
        oasysgui.lineEdit(left_box_11, self, "plot_Kmin", "Minimum K", tooltip="plot_Kmin",
                          labelWidth=250, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(left_box_11, self, "plot_Kmax", "Maximum K", tooltip="plot_Kmax",
                          labelWidth=250, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(left_box_11, self, "plot_max_harmonic_number", "Maximum harmonic number", tooltip="plot_max_harmonic_number",
                          labelWidth=250, valueType=int, orientation="horizontal")


        # undulator plots
        self.add_specific_undulator_plots()

        self.__is_init = False

        self.set_visibility()
        orangegui.rubber(self.controlArea)
        orangegui.rubber(self.mainArea)


    
    def set_TypeOfProperties(self):
        super(OWUndulatorGaussian, self).set_TypeOfProperties()

        if not self.__is_init: self.set_visibility_energy_spread()
    
    def set_visibility_energy_spread(self): # to be filled in the upper class
        self.box_energy_spread_local.setVisible(self.flag_energy_spread == 1)

    def set_visibility(self):
        self.box_flux_central_cone.setVisible(self.flag_autoset_flux_central_cone == 0)
        self.set_visibility_energy_spread()

    def add_specific_undulator_plots(self):
        undulator_plot_tab = oasysgui.widgetBox(self.main_tabs, addToLayout=0, margin=4)

        self.main_tabs.insertTab(1, undulator_plot_tab, "Undulator Plots")

        view_box = oasysgui.widgetBox(undulator_plot_tab, "Plotting Style", addSpace=False, orientation="horizontal")
        view_box_1 = oasysgui.widgetBox(view_box, "", addSpace=False, orientation="vertical", width=350)

        self.undulator_view_type_combo = orangegui.comboBox(view_box_1, self,
                                            "plot_undulator_graph",
                                                          label="Plot Graphs?",
                                                          labelWidth=220,
                                                          items=["No", "Yes"],
                                                          callback=self.refresh_specific_plots,
                                                          sendSelectedValue=False,
                                                          orientation="horizontal")

        self.undulator_tab = []
        self.undulator_tabs = oasysgui.tabWidget(undulator_plot_tab)

        current_tab = self.undulator_tabs.currentIndex()

        size = len(self.undulator_tab)
        indexes = range(0, size)
        for index in indexes:
            self.undulator_tabs.removeTab(size-1-index)

        self.undulator_tab = [
            orangegui.createTabPage(self.undulator_tabs, "Flux spectrum"),
            orangegui.createTabPage(self.undulator_tabs, "Spectral Power"),
            orangegui.createTabPage(self.undulator_tabs, "photon beam sizes"),
            orangegui.createTabPage(self.undulator_tabs, "photon beam divergences"),
        ]

        self.undulator_plot_canvas = [None, None, None, None]

        for tab in self.undulator_tab:
            tab.setFixedHeight(self.IMAGE_HEIGHT)
            tab.setFixedWidth(self.IMAGE_WIDTH)

        self.undulator_tabs.setCurrentIndex(current_tab)

    def plot_undulator_item1D(self, undulator_plot_slot_index, x, y, title="", xtitle="", ytitle="", symbol='.'):
        self.undulator_tab[undulator_plot_slot_index].layout().removeItem(self.undulator_tab[undulator_plot_slot_index].layout().itemAt(0))
        plot_widget_id = plot_data1D(x.copy(), y.copy(), title=title, xtitle=xtitle, ytitle=ytitle, symbol=symbol)
        self.undulator_tab[undulator_plot_slot_index].layout().addWidget(plot_widget_id)

    def plot_undulator_multi_data1D(self, undulator_plot_slot_index, x, y, title="", xtitle="", ytitle="", ytitles=[""], symbol='.'):
        self.undulator_tab[undulator_plot_slot_index].layout().removeItem(self.undulator_tab[undulator_plot_slot_index].layout().itemAt(0))
        plot_widget_id = plot_multi_data1D(x, y, title=title, xtitle=xtitle, ytitle=ytitle, ytitles=ytitles, flag_common_abscissas=0) #, symbol=symbol)
        self.undulator_tab[undulator_plot_slot_index].layout().addWidget(plot_widget_id)

    def refresh_specific_plots(self):
        if self.light_source is None: return

        if self.plot_undulator_graph == 0:
            for undulator_plot_slot_index in range(len(self.undulator_plot_canvas)):
                current_item = self.undulator_tab[undulator_plot_slot_index].layout().itemAt(0)
                self.undulator_tab[undulator_plot_slot_index].layout().removeItem(current_item)
                plot_widget_id = oasysgui.QLabel() # TODO: is there a better way to clean this??????????????????????
                self.undulator_tab[undulator_plot_slot_index].layout().addWidget(plot_widget_id)
        else:
            # spectra
            e, f, w = self.light_source.calculate_spectrum()
            self.plot_undulator_item1D(0, e, f,
                                  title="Undulator spectrum", xtitle="Photon energy [eV]", ytitle=r"Photons/s/0.1%bw")

            self.plot_undulator_item1D(1, e, w,
                                  title="Undulator spectral power",
                                  xtitle="Photon energy [eV]", ytitle="Spectral power [W/eV]")

            # size and divergence
            Energies, SizeH, SizeV, DivergenceH, DivergenceV, Labels = \
            self.light_source.get_size_and_divergence_vs_photon_energy(self.plot_Kmin, self.plot_Kmax,
                                                                      max_harmonic_number=self.plot_max_harmonic_number)

            Energies2 = numpy.concatenate((Energies, Energies))
            Sizes = numpy.concatenate((SizeH, SizeV))
            Divergences = numpy.concatenate((DivergenceH, DivergenceV))
            Labels2 = ["H " + item for item in Labels] + ["V " + item for item in Labels]

            self.plot_undulator_multi_data1D(2, Energies2, 1e6 * Sizes,
                                  title="Undulator size", xtitle="Photon energy [eV]", ytitle=r"Size (Sigma) [um]",
                                  ytitles=Labels2)

            self.plot_undulator_multi_data1D(3, Energies2, 1e6 * Divergences,
                                  title="Undulator divergence", xtitle="Photon energy [eV]", ytitle=r"Divergence angle (Sigma) [urad]",
                                  ytitles=Labels2)

    def check_magnetic_structure(self):
        self.energy = congruence.checkPositiveNumber(self.energy, "Energy")
        self.delta_e = congruence.checkPositiveNumber(self.delta_e, "Delta Energy")
        self.undulator_length = congruence.checkPositiveNumber(self.undulator_length, "Undulator Length")


    def build_light_source(self, electron_beam, flag_emittance):
        sourceundulator = S4UndulatorGaussian(
            period_length=self.period_length,  # syned Undulator parameter
            number_of_periods=self.undulator_length / self.period_length,  # syned Undulator parameter
            photon_energy=self.energy,
            delta_e=self.delta_e,
            ng_e=self.plot_npoints,  # Photon energy scan number of points
            flag_emittance=flag_emittance,  # when sampling rays: Use emittance (0=No, 1=Yes)
            flag_energy_spread=self.flag_energy_spread,
            harmonic_number=self.harmonic_number,
            flag_autoset_flux_central_cone=self.flag_autoset_flux_central_cone,
            flux_central_cone=float(self.flux_central_cone),
        )

        # S4UndulatorLightSource
        try:    name = self.getNode().title
        except: name = "Undulator gaussian"

        lightsource = S4UndulatorGaussianLightSource(name=name,
                                             electron_beam=electron_beam,
                                             magnetic_structure=sourceundulator,
                                             nrays=self.number_of_rays,
                                             seed=self.seed)

        print("\n\n***** S4UndulatorLightSource info: ", lightsource.info())

        return lightsource

    def populate_fields_from_magnetic_structure(self, magnetic_structure, electron_beam):
        if isinstance(magnetic_structure, Undulator):
            self.energy           = round(magnetic_structure.resonance_energy(electron_beam.gamma()))
            self.delta_e          = 0.0
            self.undulator_length = magnetic_structure.length()
            self.period_length    = magnetic_structure.period_length()
            self.plot_Kmax        = magnetic_structure.K_vertical()


add_widget_parameters_to_module(__name__)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------- #
# Copyright (c) 2025, UChicago Argonne, LLC. All rights reserved.         #
#                                                                         #
# Copyright 2025. UChicago Argonne, LLC. This software was produced       #
# under U.S. Government contract DE-AC02-06CH11357 for Argonne National   #
# Laboratory (ANL), which is operated by UChicago Argonne, LLC for the    #
# U.S. Department of Energy. The U.S. Government has rights to use,       #
# reproduce, and distribute this software.  NEITHER THE GOVERNMENT NOR    #
# UChicago Argonne, LLC MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR        #
# ASSUMES ANY LIABILITY FOR THE USE OF THIS SOFTWARE.  If software is     #
# modified to produce derivative works, such modified software should     #
# be clearly marked, so as not to confuse it with the version available   #
# from ANL.                                                               #
#                                                                         #
# Additionally, redistribution and use in source and binary forms, with   #
# or without modification, are permitted provided that the following      #
# conditions are met:                                                     #
#                                                                         #
#     * Redistributions of source code must retain the above copyright    #
#       notice, this list of conditions and the following disclaimer.     #
#                                                                         #
#     * Redistributions in binary form must reproduce the above copyright #
#       notice, this list of conditions and the following disclaimer in   #
#       the documentation and/or other materials provided with the        #
#       distribution.                                                     #
#                                                                         #
#     * Neither the name of UChicago Argonne, LLC, Argonne National       #
#       Laboratory, ANL, the U.S. Government, nor the names of its        #
#       contributors may be used to endorse or promote products derived   #
#       from this software without specific prior written permission.     #
#                                                                         #
# THIS SOFTWARE IS PROVIDED BY UChicago Argonne, LLC AND CONTRIBUTORS     #
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT       #
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS       #
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL UChicago     #
# Argonne, LLC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,        #
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,    #
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;        #
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER        #
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT      #
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN       #
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE         #
# POSSIBILITY OF SUCH DAMAGE.                                             #
# ----------------------------------------------------------------------- #
import sys
import time

from orangewidget.widget import Output
from orangewidget.settings import Setting

from oasys2.widget.util.widget_util import EmittingStream
from oasys2.widget.util import congruence

from syned.widget.widget_decorator import WidgetDecorator
from syned.beamline.beamline import Beamline

from orangecontrib.shadow4.widgets.gui.ow_electron_beam import OWElectronBeam
from orangecontrib.shadow4.util.shadow4_objects import ShadowData
from orangecontrib.shadow4.util.shadow4_util import TriggerToolsDecorator, TriggerIn

from shadow4.tools.logger import set_verbose
from shadow4.beamline.s4_beamline import S4Beamline

class OWSynchrotronSource(OWElectronBeam, WidgetDecorator, TriggerToolsDecorator):
    class Inputs:
        trigger     = TriggerToolsDecorator.get_trigger_input()
        syned_data  = WidgetDecorator.syned_input_data(multi_input=True)

    class Outputs:
        shadow_data = Output("Shadow Data", ShadowData, default=True, auto_summary=False)
        trigger     = TriggerToolsDecorator.get_trigger_output()

    # sampling rays
    number_of_rays = Setting(500)
    seed           = Setting(5676561)

    light_source = None

    def __init__(self, show_energy_spread=False):
        super().__init__(show_energy_spread=show_energy_spread)

    @Inputs.trigger
    def set_trigger_parameters_for_sources(self, trigger):
        super(OWSynchrotronSource, self).set_trigger_parameters_for_sources(trigger)

    @Inputs.syned_data
    def set_syned_data(self, index, syned_data):
        self.receive_syned_data(syned_data)

    @Inputs.syned_data.insert
    def insert_syned_data(self, index, syned_data):
        self.receive_syned_data(syned_data)

    @Inputs.syned_data.remove
    def remove_syned_data(self, index):
        pass

    def receive_syned_data(self, data):
        if data is not None:
            if isinstance(data, Beamline):
                light_source = data.get_light_source()
                if light_source is not None:
                    electron_beam      = light_source.get_electron_beam()
                    magnetic_structure = light_source.get_magnetic_structure()

                    if electron_beam is not None:      self.populate_fields_from_electron_beam(electron_beam) # from OWElectronBeam
                    if magnetic_structure is not None: self.populate_fields_from_magnetic_structure(magnetic_structure, electron_beam)

                    self.type_of_properties = 2 if self._check_dispersion_presence() else 1
                    self.set_TypeOfProperties()
                else:
                    raise ValueError("Syned data not correct: light source not present")
            else:
                raise ValueError("Syned data not correct: it must be Beamline()")

    def check_data(self):
        self.number_of_rays = congruence.checkPositiveNumber(self.number_of_rays, "Number of rays")
        self.seed           = congruence.checkPositiveNumber(self.seed, "Seed")

        self.check_electron_beam() # from OWElectronBeam
        self.check_magnetic_structure()

    def get_light_source(self):
        electron_beam = self.get_electron_beam() # from OWElectronBeam

        if not electron_beam is None: # None if user canceled operation
            print("\n\n***** electron_beam info: ", electron_beam.info())

            if self.type_of_properties == 3: flag_emittance = 0
            else:                            flag_emittance = 1

            return self.build_light_source(electron_beam, flag_emittance)
        else:
            return None

    def run_shadow4(self, scanning_data: ShadowData.ScanningData = None):
        try:
            if not scanning_data: scanning_data = None

            self.check_data()

            light_source = self.get_light_source()

            if not light_source is None: # None if user has canceled the operation
                self.light_source = None
                set_verbose()
                self.shadow_output.setText("")
                sys.stdout = EmittingStream(textWritten=self._write_stdout)

                self._set_plot_quality()

                self.progressBarInit()

                #
                # script
                #
                script = light_source.to_python_code()
                script += "\n\n# test plot\nfrom srxraylib.plot.gol import plot_scatter"
                script += "\nrays = beam.get_rays()"
                script += "\nplot_scatter(1e6 * rays[:, 0], 1e6 * rays[:, 2], title='(X,Z) in microns')"
                self.shadow4_script.set_code(script)
                self.progressBarSet(5)

                # run shadow4
                t00 = time.time()
                print("***** starting calculation...")
                output_beam = light_source.get_beam()
                t11 = time.time() - t00
                print("***** time for %d rays: %f s, %f min, " % (self.number_of_rays, t11, t11 / 60))

                self.light_source = light_source

                #
                # plots
                #
                self._plot_results(output_beam, None, progressBarValue=80)
                self.refresh_specific_plots()


                #
                # send beam and trigger
                #
                output_data = ShadowData(beam=output_beam,
                                         number_of_rays=self.number_of_rays,
                                         beamline=S4Beamline(light_source=light_source))
                output_data.scanning_data = scanning_data

                self.Outputs.shadow_data.send(output_data)
                self.Outputs.trigger.send(TriggerIn(new_object=True))
        except Exception as exception:
            try:    self._initialize_tabs()
            except: pass
            self.prompt_exception(exception)
        finally:
            self.progressBarFinished()

    def build_light_source(self, electron_beam, flag_emittance): raise NotImplementedError
    def populate_fields_from_magnetic_structure(self, magnetic_structure, electron_beam): raise NotImplementedError
    def check_magnetic_structure(self): raise NotImplementedError
    def refresh_specific_plots(self): raise NotImplementedError



import os
import pickle

from AnyQt.QtWidgets import QMessageBox

from orangewidget import gui
from orangewidget.settings import Setting
from orangewidget.widget import Input, Output

from oasys2.widget import gui as oasysgui
from oasys2.widget.util import congruence
from oasys2.widget.widget import OWWidget, OWAction
from oasys2.canvas.util.canvas_util import add_widget_parameters_to_module

from orangecontrib.shadow4.util.shadow4_objects import ShadowData
from orangecontrib.shadow4.util.shadow4_util import ShadowCongruence

class BeamFileWriter(OWWidget):
    name = "Shadow4 File Writer"
    description = "Tools: Shadow4 File Writer"
    icon = "icons/beam_file_writer.png"
    maintainer = "Luca Rebuffi"
    maintainer_email = "lrebuffi(@at@)anl.gov"
    priority = 7
    category = "Tools"
    keywords = ["data", "file", "load", "read"]

    want_main_area = 0

    shadow_data_file_name = Setting("")
    is_automatic_run      = Setting(1)

    class Inputs:
        shadow_data = Input("Shadow Data", ShadowData, default=True, auto_summary=False)

    class Outputs:
        shadow_data = Output("Shadow Data", ShadowData, default=True, auto_summary=False)


    input_data = None

    def __init__(self):
        super().__init__()

        self.runaction = OWAction("Write Shadow4 File", self)
        self.runaction.triggered.connect(self.write_file)
        self.addAction(self.runaction)

        self.setFixedWidth(590)
        self.setFixedHeight(210)

        left_box_1 = oasysgui.widgetBox(self.controlArea, "Shadow4 File Selection", addSpace=True, orientation="vertical",
                                         width=570, height=110)

        gui.checkBox(left_box_1, self, 'is_automatic_run', 'Automatic Execution')

        gui.separator(left_box_1, height=10)

        figure_box = oasysgui.widgetBox(left_box_1, "", addSpace=True, orientation="horizontal", width=550, height=35)

        self.le_shadow_data_file_name = oasysgui.lineEdit(figure_box, self, "shadow_data_file_name", "Shadow4 File Name",
                                                    labelWidth=120, valueType=str, orientation="horizontal")
        self.le_shadow_data_file_name.setFixedWidth(330)

        gui.button(figure_box, self, "...", callback=self.selectFile)

        button = gui.button(self.controlArea, self, "Write Shadow4 File", callback=self.write_file)
        button.setFixedHeight(45)
        button.setFixedWidth(570)

        gui.rubber(self.controlArea)

    def selectFile(self):
        self.le_shadow_data_file_name.setText(oasysgui.selectSaveFileFromDialog(self, self.shadow_data_file_name, default_file_name="s4_data.pkl", file_extension_filter="Pickle Files (*.pkl)"))

    @Inputs.shadow_data
    def set_shadow_data(self, input_data: ShadowData):
        if ShadowCongruence.check_empty_data(input_data):
            if ShadowCongruence.check_good_beam(input_data.beam):
                self.input_data = input_data
            else:
                QMessageBox.critical(self, "Error", "No good rays or bad content", QMessageBox.Ok)
                return
        else:
            QMessageBox.critical(self, "Error", "Empty input data or empty beam", QMessageBox.Ok)

        if self.is_automatic_run: self.write_file()

    def write_file(self):
        self.setStatusMessage("")

        try:
            if ShadowCongruence.check_empty_data(self.input_data):
                if congruence.checkFileName(self.shadow_data_file_name):
                    with open(self.shadow_data_file_name, "wb") as f:  pickle.dump(self.input_data, f)

                    _, file_name = os.path.split(self.shadow_data_file_name)

                    self.setStatusMessage("Current: " + file_name)

                    self.Outputs.shadow_data.send(self.input_data)
            else:
                QMessageBox.critical(self, "Error", "Empty input data or empty beam", QMessageBox.Ok)
        except Exception as exception:
            QMessageBox.critical(self, "Error", str(exception), QMessageBox.Ok)

add_widget_parameters_to_module(__name__)




import os, copy, numpy
from shadow4.beam.s4_beam import S4Beam

class ShadowData:
    class ScanningData(object):
        def __init__(self,
                     scanned_variable_name,
                     scanned_variable_value,
                     scanned_variable_display_name,
                     scanned_variable_um,
                     additional_parameters={}):
            self.__scanned_variable_name = scanned_variable_name
            self.__scanned_variable_value = scanned_variable_value
            self.__scanned_variable_display_name = scanned_variable_display_name
            self.__scanned_variable_um = scanned_variable_um
            self.__additional_parameters=additional_parameters

        @property
        def scanned_variable_name(self):
            return self.__scanned_variable_name

        @property
        def scanned_variable_value(self):
            return self.__scanned_variable_value

        @property
        def scanned_variable_display_name(self):
            return self.__scanned_variable_display_name

        @property
        def scanned_variable_um(self):
            return self.__scanned_variable_um

        def has_additional_parameter(self, name):
            return name in self.__additional_parameters.keys()

        def get_additional_parameter(self, name):
            return self.__additional_parameters[name]

    def __init__(self, beam=None, footprint=None, number_of_rays=0, beamline=None):
        if (beam is None):
            if number_of_rays > 0: self.__beam = S4Beam(number_of_rays)
            else:                  self.__beam = S4Beam()
            self.__footprint = None
        else:
            self.__beam      = beam
            self.__footprint = footprint

        self.__scanning_data = None
        self.__initial_flux  = None
        self.__beamline      = beamline  # added by srio

    @property
    def beam(self):
        return self.__beam

    @beam.setter
    def beam(self, beam):
        self.__beam = beam

    @property
    def footprint(self):
        return self.__footprint

    @footprint.setter
    def footprint(self, footprint):
        self._footprint = footprint

    @property
    def beamline(self):
        return self.__beamline

    @beamline.setter
    def beamline(self, beamline):
        self.__beamline = beamline

    @property
    def initial_flux(self):
        return self.__initial_flux

    @initial_flux.setter
    def initial_flux(self, initial_flux):
        self.__initial_flux = initial_flux

    @property
    def scanning_data(self):
        return self.__scanning_data

    @scanning_data.setter
    def scanning_data(self, scanning_data : ScanningData):
        self.__scanning_data = scanning_data

    def get_flux(self, nolost=1):
        if not self.__beam is None and not self.__initial_flux is None:
            return (self.__beam.intensity(nolost) / self.get_number_of_rays(0)) * self.get_initial_flux()
        else:
            return None

    def get_number_of_rays(self, nolost=0):
        if not hasattr(self.__beam, "rays"): return 0
        if nolost == 0:   return self.__beam.rays.shape[0]
        elif nolost == 1: return self.__beam.rays[numpy.where(self.__beam.rays[:, 9] > 0)].shape[0]
        elif nolost == 2: return self.__beam.rays[numpy.where(self.__beam.rays[:, 9] < 0)].shape[0]
        else: raise ValueError("nolost flag value not valid")

    def load_from_file(self, file_name):
        if not self.__beam is None:
            if os.path.exists(file_name): self.__beam.load_h5(file_name)
            else: raise Exception("File " + file_name + " not existing")

    def write_to_file(self, file_name):
        if not self.__beam is None:
            self.__beam.write_h5(file_name)

    def duplicate(self, copy_rays=True, copy_beamline=True):
        beam      = S4Beam()
        footprint = None if self.__footprint is None else S4Beam()

        if copy_rays:
            beam.rays = copy.deepcopy(self.beam.rays)
            if not self.footprint is None:
                footprint = S4Beam()
                footprint.rays = copy.deepcopy(self.footprint.rays)

        new_shadow_beam = ShadowData(beam=beam,
                                     footprint=footprint)

        new_shadow_beam.scanning_data = self.__scanning_data
        new_shadow_beam.initial_flux  = self.__initial_flux

        if copy_beamline: new_shadow_beam.beamline = self.__beamline.duplicate()

        return new_shadow_beam

    @classmethod
    def merge_shadow_data(cls, data_1, data_2, which_flux=3, which_beamline=0):
        if data_1 and data_2:
            data_1: ShadowData = data_1
            data_2: ShadowData = data_2

            has_footprint = not (data_1.footprint is None or data_2.footprint is None)

            rays_1 = None
            rays_2 = None
            footprint_rays_1 = None
            footprint_rays_2 = None

            if len(getattr(data_1.beam, "rays", numpy.zeros(0))) > 0: rays_1 = copy.deepcopy(data_1.beam.rays)
            if len(getattr(data_2.beam, "rays", numpy.zeros(0))) > 0: rays_2 = copy.deepcopy(data_2.beam.rays)
            if has_footprint:
                if len(getattr(data_1.footprint, "rays", numpy.zeros(0))) > 0: footprint_rays_1 = copy.deepcopy(data_1.footprint.rays)
                if len(getattr(data_2.footprint, "rays", numpy.zeros(0))) > 0: footprint_rays_2 = copy.deepcopy(data_2.footprint.rays)

            merged_beam : ShadowData = data_1.duplicate(copy_rays=False, copy_beamline=False)

            merged_beam.beam.rays        = numpy.append(rays_1, rays_2, axis=0)
            merged_beam.beam.rays[:, 11] = numpy.arange(1, len(merged_beam.beam.rays) + 1, 1) # ray_index

            if has_footprint:
                merged_beam.footprint.rays        = numpy.append(footprint_rays_1, footprint_rays_2, axis=0)
                merged_beam.footprint.rays[:, 11] = numpy.arange(1, len(merged_beam.footprint.rays) + 1, 1)  # ray_index

            if which_flux ==1 :
                if not data_1.initial_flux is None:
                    merged_beam.initial_flux = data_1.initial_flux
            elif which_flux == 2:
                if not data_2.initial_flux is None:
                    merged_beam.initial_flux = data_2.initial_flux
            else:
                if not data_1.initial_flux is None and not data_2.initial_flux is None:
                    merged_beam.initial_flux = data_1.initial_flux + data_2.initial_flux

            if which_beamline == 0:   merged_beam.beamline = data_1.beamline
            elif which_beamline == 1: merged_beam.beamline = data_2.beamline
            else:                     merged_beam.beamline = None

            return merged_beam
        else:
            raise Exception("Both input beams should provided for merging")

    @classmethod
    def initialize_from_beam(cls, input_beam):
        return input_beam.duplicate()


# TODO: review all preprocessor data

class MLayerPreProcessorData:
    NONE = "None"
    def __init__(self,
                 mlayer_data_file=NONE,
                 ):
        super().__init__()
        self.mlayer_data_file = mlayer_data_file

class PreReflPreProcessorData:
    NONE = "None"
    def __init__(self,
                 prerefl_data_file=NONE,
                 ):
        super().__init__()
        self.prerefl_data_file = prerefl_data_file


class BraggPreProcessorData:
    NONE = "None"
    def __init__(self,
                 bragg_data_file=NONE):
        super().__init__()

        self.bragg_data_file = bragg_data_file


class VlsPgmPreProcessorData:
    def __init__(self,
                 shadow_coeff_0=0.0,
                 shadow_coeff_1=0.0,
                 shadow_coeff_2=0.0,
                 shadow_coeff_3=0.0,
                 d_source_plane_to_mirror=0.0,
                 d_mirror_to_grating=0.0,
                 d_grating_to_exit_slits=0.0,
                 alpha=0.0,
                 beta=0.0):
        self.shadow_coeff_0 = shadow_coeff_0
        self.shadow_coeff_1 = shadow_coeff_1
        self.shadow_coeff_2 = shadow_coeff_2
        self.shadow_coeff_3 = shadow_coeff_3
        self.d_source_plane_to_mirror = d_source_plane_to_mirror
        self.d_mirror_to_grating = d_mirror_to_grating
        self.d_grating_to_exit_slits = d_grating_to_exit_slits
        self.alpha = alpha
        self.beta = beta




import copy
import numpy

from orangewidget import gui
from orangewidget.widget import Input, Output

from orangewidget.settings import Setting
from oasys2.widget import gui as oasysgui
from oasys2.widget.gui import ConfirmDialog, Styles
from oasys2.widget.util.widget_objects import TriggerIn
from oasys2.canvas.util.canvas_util import add_widget_parameters_to_module

from orangecontrib.shadow4.util.shadow4_objects import ShadowData, S4Beam
from orangecontrib.shadow4.util.shadow4_util import ShadowCongruence
from orangecontrib.shadow4.widgets.gui.ow_automatic_element import AutomaticElement
from orangecontrib.shadow4.util.shadow4_util import TriggerToolsDecorator

class AccumulatingLoopPoint(AutomaticElement):
    name = "Beam Accumulating Point"
    description = "User Defined: Beam Accumulating Point"
    icon = "icons/beam_accumulating.png"
    maintainer = "Luca Rebuffi"
    maintainer_email = "lrebuffi(@at@)anl.gov"
    priority = 2
    category = "User Defined"
    keywords = ["data", "file", "load", "read"]

    class Inputs:
        shadow_data = Input("Shadow Data", ShadowData, default=True, auto_summary=False)

    class Outputs:
        shadow_data = Output("Shadow Data", ShadowData, default=True, auto_summary=False)
        trigger     = TriggerToolsDecorator.get_trigger_output()

    IMAGE_WIDTH = 878
    IMAGE_HEIGHT = 635

    input_data : ShadowData = None

    want_main_area = 0

    number_of_accumulated_rays = Setting(10000)
    kind_of_accumulation = Setting(0)

    current_number_of_rays = 0
    current_intensity = 0
    current_number_of_total_rays = 0
    current_number_of_lost_rays = 0

    keep_go_rays = Setting(1)

    def __init__(self):
        super().__init__(show_automatic_box=False)
        self.is_automatic_run = True

        self.setFixedWidth(570)
        self.setFixedHeight(380)

        self.controlArea.setFixedWidth(560)

        button_box = gui.widgetBox(self.controlArea, "", addSpace=True, orientation="horizontal")

        self.start_button = gui.button(button_box, self, "Send Beam", callback=self.send_signal)
        self.start_button.setFixedHeight(45)

        button = gui.button(button_box, self, "Reset Accumulation", callback=self.callResetSettings)
        button.setStyleSheet(Styles.button_blue)

        left_box_1 = oasysgui.widgetBox(self.controlArea, "Accumulating Loop Management", addSpace=False, orientation="vertical", height=260)

        gui.comboBox(left_box_1, self, "kind_of_accumulation", label="Accumulated Quantity", labelWidth=350,
                     items=["Number of Good Rays ", "Intensity of Good Rays"],
                     callback=self.set_KindOfAccumulation,
                     sendSelectedValue=False, orientation="horizontal")

        self.left_box_1_1 = oasysgui.widgetBox(left_box_1, "", addSpace=False, orientation="vertical", height=35)
        self.left_box_1_2 = oasysgui.widgetBox(left_box_1, "", addSpace=False, orientation="vertical", height=35)

        oasysgui.lineEdit(self.left_box_1_1, self, "number_of_accumulated_rays", "Number of accumulated good rays\n(before sending signal)", labelWidth=350, valueType=float,
                           orientation="horizontal")

        oasysgui.lineEdit(self.left_box_1_2, self, "number_of_accumulated_rays", "Intenisty of accumulated good rays\n(before sending signal)", labelWidth=350, valueType=float,
                           orientation="horizontal")

        self.set_KindOfAccumulation()

        gui.comboBox(left_box_1, self, "keep_go_rays", label="Remove lost rays from beam", labelWidth=350, items=["No", "Yes"], sendSelectedValue=False, orientation="horizontal")

        gui.separator(left_box_1)

        le = oasysgui.lineEdit(left_box_1, self, "current_number_of_rays", "Current number of good rays", labelWidth=350, valueType=int, orientation="horizontal")
        le.setReadOnly(True)
        le.setStyleSheet(Styles.line_edit_read_only)

        self.le_current_intensity = oasysgui.lineEdit(left_box_1, self, "current_intensity", "Current intensity", labelWidth=350, valueType=float, orientation="horizontal")
        self.le_current_intensity.setReadOnly(True)
        self.le_current_intensity.setStyleSheet(Styles.line_edit_read_only)

        le = oasysgui.lineEdit(left_box_1, self, "current_number_of_lost_rays", "Current number of lost rays", labelWidth=350, valueType=int, orientation="horizontal")
        le.setReadOnly(True)
        le.setStyleSheet("color: darkred; background-color: rgb(243, 240, 160);")

        le = oasysgui.lineEdit(left_box_1, self, "current_number_of_total_rays", "Current number of total rays", labelWidth=350, valueType=int, orientation="horizontal")
        le.setReadOnly(True)
        le.setStyleSheet("color: black; background-color: rgb(243, 240, 160);")

        gui.rubber(self.controlArea)

    def set_KindOfAccumulation(self):
        self.left_box_1_1.setVisible(self.kind_of_accumulation==0)
        self.left_box_1_2.setVisible(self.kind_of_accumulation==1)

    def send_signal(self):
        self.Outputs.shadow_data.send(self.input_data)
        self.Outputs.trigger.send(TriggerIn(interrupt=True))

    def callResetSettings(self):
        if ConfirmDialog.confirmed(parent=self, message="Confirm Reset of the accumulated beam"):
            self.current_number_of_rays = 0
            self.current_intensity = 0.0
            self.current_number_of_lost_rays = 0
            self.current_number_of_total_rays = 0
            self.input_data = None

    @Inputs.shadow_data
    def set_shadow_data(self, input_data: ShadowData):
        if ShadowCongruence.check_empty_data(input_data):
            proceed = True
            beam : S4Beam     = input_data.beam
            footprint: S4Beam = input_data.footprint

            if not ShadowCongruence.check_good_beam(beam):
                if not ConfirmDialog.confirmed(parent=self, message="Beam contains bad values, skip it?"):
                    proceed = False

            if proceed:
                scanning_data = input_data.scanning_data

                go = numpy.where(beam.rays[:, 9] == 1)

                nr_good  = len(beam.rays[go])
                nr_total = len(beam.rays)
                nr_lost  = nr_total - nr_good

                intensity = beam.histo1(1, nolost=1, ref=23)['intensity']

                self.current_number_of_rays       += nr_good
                self.current_intensity            += intensity
                self.current_number_of_lost_rays  += nr_lost
                self.current_number_of_total_rays += nr_total

                self.le_current_intensity.setText("{:10.3f}".format(self.current_intensity))

                if self.keep_go_rays == 1:
                    beam.rays = copy.deepcopy(beam.rays[go])
                    if not footprint is None: footprint.rays = copy.deepcopy(footprint.rays[go])

                if not self.input_data is None:
                    self.input_data = ShadowData.merge_shadow_data(self.input_data, input_data, which_flux=3, which_beamline=0)
                else:
                    beam.rays[:, 11] = numpy.arange(1, len(beam.rays) + 1, 1)  # ray_index
                    if not footprint is None: footprint.rays[:, 11] = numpy.arange(1, len(footprint.rays) + 1, 1)

                    self.input_data = input_data

                self.input_data.scanning_data = scanning_data

                if (self.kind_of_accumulation == 0 and self.current_number_of_rays < self.number_of_accumulated_rays) or \
                   (self.kind_of_accumulation == 1 and self.current_intensity < self.number_of_accumulated_rays):
                    self.Outputs.trigger.send(TriggerIn(new_object=True))
                else:
                    self.send_signal()

                    self.current_number_of_rays       = 0
                    self.current_intensity            = 0.0
                    self.current_number_of_lost_rays  = 0
                    self.current_number_of_total_rays = 0

                    self.input_data = None

add_widget_parameters_to_module(__name__)


import os
import pickle

from AnyQt.QtWidgets import QMessageBox

from orangewidget import gui
from orangewidget.settings import Setting
from orangewidget.widget import Output

from oasys2.widget import gui as oasysgui
from oasys2.widget.util import congruence
from oasys2.widget.widget import OWWidget, OWAction
from oasys2.canvas.util.canvas_util import add_widget_parameters_to_module

from orangecontrib.shadow4.util.shadow4_objects import ShadowData


class BeamFileReader(OWWidget):
    name = "Shadow4 File Reader"
    description = "Tools: Shadow File Reader"
    icon = "icons/beam_file_reader.png"
    maintainer = "Luca Rebuffi"
    maintainer_email = "lrebuffi(@at@)anl.gov"
    priority = 6
    category = "Tools"
    keywords = ["data", "file", "load", "read"]

    want_main_area = 0

    shadow_data_file_name = Setting("")

    class Outputs:
        shadow_data = Output("Shadow Data", ShadowData, default=True, auto_summary=False)

    def __init__(self):
        super().__init__()

        self.runaction = OWAction("Read Shadow4 File", self)
        self.runaction.triggered.connect(self.read_file)
        self.addAction(self.runaction)

        self.setFixedWidth(590)
        self.setFixedHeight(170)

        left_box_1 = oasysgui.widgetBox(self.controlArea, "Shadow4 File Selection", addSpace=True, orientation="vertical",
                                         width=570, height=70)

        figure_box = oasysgui.widgetBox(left_box_1, "", addSpace=True, orientation="horizontal", width=550, height=35)

        self.le_shadow_data_file_name = oasysgui.lineEdit(figure_box, self, "shadow_data_file_name", "Shadow4 File Name",
                                                    labelWidth=120, valueType=str, orientation="horizontal")
        self.le_shadow_data_file_name.setFixedWidth(330)

        gui.button(figure_box, self, "...", callback=self.selectFile)

        button = gui.button(self.controlArea, self, "Read Shadow4 File", callback=self.read_file)
        button.setFixedHeight(45)
        button.setFixedWidth(570)

        gui.rubber(self.controlArea)

    def selectFile(self):
        self.le_shadow_data_file_name.setText(oasysgui.selectFileFromDialog(self, self.shadow_data_file_name, "Open Shadow4 File", file_extension_filter="Pickle Files (*.pkl)"))

    def read_file(self):
        self.setStatusMessage("")

        try:
            if congruence.checkFileName(self.shadow_data_file_name):
                with open(self.shadow_data_file_name, "rb") as f:  shadow_data = pickle.load(f)

                _, file_name = os.path.split(self.shadow_data_file_name)

                self.setStatusMessage("Current: " + file_name)

                self.Outputs.shadow_data.send(shadow_data)
        except Exception as exception:
            QMessageBox.critical(self, "Error", str(exception), QMessageBox.Ok)

add_widget_parameters_to_module(__name__)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# #########################################################################
# Copyright (c) 2020, UChicago Argonne, LLC. All rights reserved.         #
#                                                                         #
# Copyright 2020. UChicago Argonne, LLC. This software was produced       #
# under U.S. Government contract DE-AC02-06CH11357 for Argonne National   #
# Laboratory (ANL), which is operated by UChicago Argonne, LLC for the    #
# U.S. Department of Energy. The U.S. Government has rights to use,       #
# reproduce, and distribute this software.  NEITHER THE GOVERNMENT NOR    #
# UChicago Argonne, LLC MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR        #
# ASSUMES ANY LIABILITY FOR THE USE OF THIS SOFTWARE.  If software is     #
# modified to produce derivative works, such modified software should     #
# be clearly marked, so as not to confuse it with the version available   #
# from ANL.                                                               #
#                                                                         #
# Additionally, redistribution and use in source and binary forms, with   #
# or without modification, are permitted provided that the following      #
# conditions are met:                                                     #
#                                                                         #
#     * Redistributions of source code must retain the above copyright    #
#       notice, this list of conditions and the following disclaimer.     #
#                                                                         #
#     * Redistributions in binary form must reproduce the above copyright #
#       notice, this list of conditions and the following disclaimer in   #
#       the documentation and/or other materials provided with the        #
#       distribution.                                                     #
#                                                                         #
#     * Neither the name of UChicago Argonne, LLC, Argonne National       #
#       Laboratory, ANL, the U.S. Government, nor the names of its        #
#       contributors may be used to endorse or promote products derived   #
#       from this software without specific prior written permission.     #
#                                                                         #
# THIS SOFTWARE IS PROVIDED BY UChicago Argonne, LLC AND CONTRIBUTORS     #
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT       #
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS       #
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL UChicago     #
# Argonne, LLC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,        #
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,    #
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;        #
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER        #
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT      #
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN       #
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE         #
# POSSIBILITY OF SUCH DAMAGE.                                             #
# #########################################################################

import numpy, copy

from orangewidget import gui
from orangewidget.widget import Input, Output

from oasys2.canvas.util.canvas_util import add_widget_parameters_to_module
from oasys2.widget.widget import OWWidget

from orangecontrib.shadow4.util.shadow4_objects import ShadowData

from orangecontrib.shadow4.util.shadow4_util import ShadowCongruence, TriggerToolsDecorator
from oasys2.widget.util.widget_objects import TriggerIn

class BeamCleaner(OWWidget, TriggerToolsDecorator):
    name = "Beam Cleaner"
    description = "Tools: Beam Cleaner"
    icon = "icons/clean.png"
    maintainer = "Luca Rebuffi"
    maintainer_email = "lrebuffi(@at@)anl.gov"
    priority = 30
    category = "User Defined"
    keywords = ["data", "file", "load", "read"]

    class Inputs:
        shadow_data = Input("Shadow Data", ShadowData, default=True, auto_summary=False)

    class Outputs:
        shadow_data = Output("Shadow Data", ShadowData, default=True, auto_summary=False)
        trigger = TriggerToolsDecorator.get_trigger_output()

    want_main_area = 0
    want_control_area = 1

    def __init__(self):
         self.setFixedWidth(300)
         self.setFixedHeight(120)

         gui.separator(self.controlArea, height=20)
         gui.label(self.controlArea, self, "         LOST RAYS REMOVER", orientation="horizontal")
         gui.rubber(self.controlArea)

    @Inputs.shadow_data
    def set_shadow_data(self, shadow_data: ShadowData):
        if ShadowCongruence.check_empty_data(shadow_data):
            output_data = shadow_data.duplicate()

            if ShadowCongruence.check_good_beam(input_beam=output_data.beam):
                good = numpy.where(output_data.beam.rays[:, 9] == 1)
                output_data.beam.rays = copy.deepcopy(output_data.beam.rays[good])

            self.Outputs.shadow_data.send(output_data)
            self.Outputs.trigger.send(TriggerIn(new_object=True))

add_widget_parameters_to_module(__name__)


import sys

from orangewidget import gui
from orangewidget.settings import Setting
from orangewidget.widget import Output

from oasys2.widget import gui as oasysgui
from oasys2.widget.widget import OWAction
from oasys2.widget.util.widget_util import EmittingStream
from oasys2.widget.gui import Styles
from oasys2.canvas.util.canvas_util import add_widget_parameters_to_module

from orangecontrib.shadow4.util.shadow4_objects import ShadowData
from orangecontrib.shadow4.widgets.gui.ow_generic_element import GenericElement

from shadow4.beamline.s4_beamline import S4Beamline
from shadow4.sources.source_geometrical.source_grid_polar import SourceGridPolar
from shadow4.sources.source_geometrical.source_grid_cartesian import SourceGridCartesian
from shadow4.tools.logger import set_verbose

from orangecontrib.shadow4.util.shadow4_util import TriggerToolsDecorator
from oasys2.widget.util.widget_objects import TriggerIn

class OWGrid(GenericElement, TriggerToolsDecorator):
    name = "Grid Source"
    description = "Shadow Source: Grid Source"
    icon = "icons/grid.png"
    priority = 2

    class Inputs:
        trigger     = TriggerToolsDecorator.get_trigger_input()

    class Outputs:
        shadow_data = Output("Shadow Data", ShadowData, default=True, auto_summary=False)
        trigger     = TriggerToolsDecorator.get_trigger_output()

    # for both grid and cartesian
    coordinates = Setting(1)

    real_space_width_x = Setting(2e-3)
    real_space_width_z = Setting(2e-3)
    real_space_center_x = Setting(0.0)
    real_space_center_z = Setting(0.0)

    direction_space_width_x = Setting(20e-3)
    direction_space_width_z = Setting(20e-3)
    direction_space_center_x = Setting(0.0)
    direction_space_center_z = Setting(0.0)

    # for grid
    real_space_points_r = Setting(2)
    real_space_points_theta = Setting(8)

    direction_space_points_r = Setting(3)
    direction_space_points_theta = Setting(359)

    # for cartesian
    direction_space_width_y = Setting(20e-3)
    real_space_points_x = Setting(10)
    real_space_points_y = Setting(10)
    real_space_points_z = Setting(10)

    direction_space_points_x = Setting(1)
    direction_space_points_z = Setting(1)


    units=Setting(0)
    single_line_value = Setting(1000.0)

    # polarization = Setting(1)
    polarization_phase_deg = Setting(0.0)
    polarization_degree = Setting(1.0)
    coherent_beam = Setting(1)


    def __init__(self):
        super().__init__(show_automatic_box=False, has_footprint=False)

        self.runaction = OWAction("Run Shadow4/Source", self)
        self.runaction.triggered.connect(self.run_shadow4)
        self.addAction(self.runaction)

        button_box = oasysgui.widgetBox(self.controlArea, "", addSpace=False, orientation="horizontal", width=self.CONTROL_AREA_WIDTH-5)

        button = gui.button(button_box, self, "Run Shadow4/Source", callback=self.run_shadow4)
        button.setStyleSheet(Styles.button_blue)

        button = gui.button(button_box, self, "Reset Fields", callback=self.call_reset_settings)
        button.setStyleSheet(Styles.button_red)

        ################################################################################################################
        self.controlArea.setFixedWidth(self.CONTROL_AREA_WIDTH)

        tabs_setting = oasysgui.tabWidget(self.controlArea)
        tabs_setting.setFixedHeight(self.TABS_AREA_HEIGHT)
        tabs_setting.setFixedWidth(self.CONTROL_AREA_WIDTH - 5)

        tab_basic = oasysgui.createTabPage(tabs_setting, "General")
        tab_geometry = oasysgui.createTabPage(tabs_setting, "Geometry")
        tab_energy = oasysgui.createTabPage(tabs_setting, "Energy/Polarization")

        ##############################
        # BASIC

        left_box_1 = oasysgui.widgetBox(tab_basic, "Coordinates", addSpace=True, orientation="vertical")
        gui.comboBox(left_box_1, self, "coordinates", label="Coordinates", labelWidth=355,
                     items=["Cartesian", "Polar"], orientation="horizontal",
                     callback=self.set_coordinates_visibility)

        #### points
        points_box = oasysgui.widgetBox(tab_basic, "Number of points", addSpace=True,
                                                   orientation="vertical")
        ##
        self.points_polar_box = oasysgui.widgetBox(points_box, "", addSpace=True, orientation="vertical")

        oasysgui.lineEdit(self.points_polar_box, self, "real_space_points_r",
                          "Real space Radial points", tooltip="real_space_points_r", labelWidth=260, valueType=int,
                          orientation="horizontal")

        oasysgui.lineEdit(self.points_polar_box, self, "real_space_points_theta",
                          "Real Space Azimuthal points", tooltip="real_space_points_theta", labelWidth=260, valueType=int,
                          orientation="horizontal")

        oasysgui.lineEdit(self.points_polar_box, self, "direction_space_points_r",
                          "Direction space Radial points", tooltip="direction_space_points_r", labelWidth=260, valueType=int,
                          orientation="horizontal")

        oasysgui.lineEdit(self.points_polar_box, self, "direction_space_points_theta",
                          "Direction space Azimuthal points", tooltip="direction_space_points_theta", labelWidth=260, valueType=int,
                          orientation="horizontal")

        ##
        self.points_cartesian_box = oasysgui.widgetBox(points_box, "", addSpace=True, orientation="vertical")

        oasysgui.lineEdit(self.points_cartesian_box, self, "real_space_points_x",
                          "Real space X points", tooltip="real_space_points_x", labelWidth=260, valueType=int,
                          orientation="horizontal")

        oasysgui.lineEdit(self.points_cartesian_box, self, "real_space_points_z",
                          "Real Space Z points", tooltip="real_space_points_z", labelWidth=260, valueType=int,
                          orientation="horizontal")

        oasysgui.lineEdit(self.points_cartesian_box, self, "direction_space_points_x",
                          "Direction space X' points", tooltip="direction_space_points_x", labelWidth=260, valueType=int,
                          orientation="horizontal")

        oasysgui.lineEdit(self.points_cartesian_box, self, "direction_space_points_z",
                          "Direction space Z' points", tooltip="direction_space_points_z", labelWidth=260, valueType=int,
                          orientation="horizontal")
        ##############################
        # GEOMETRY

        left_box_2 = oasysgui.widgetBox(tab_geometry, "", addSpace=True, orientation="vertical", height=550)

        ###### real space
        real_distribution_box = oasysgui.widgetBox(left_box_2, "Real space", addSpace=True,
                                                      orientation="vertical") #, height=260)

        oasysgui.lineEdit(real_distribution_box, self, "real_space_center_x",
                          "Center in X [m]", tooltip="real_space_center_x", labelWidth=260, valueType=float,
                          orientation="horizontal")

        oasysgui.lineEdit(real_distribution_box, self, "real_space_center_z",
                          "Center in Z [m]", tooltip="real_space_center_z", labelWidth=260, valueType=float,
                          orientation="horizontal")

        oasysgui.lineEdit(real_distribution_box, self, "real_space_width_x",
                          "Width in X [m]", tooltip="real_space_width_x", labelWidth=260, valueType=float, orientation="horizontal")

        oasysgui.lineEdit(real_distribution_box, self, "real_space_width_z",
                          "Width in Z [m]", tooltip="real_space_width_z", labelWidth=260, valueType=float, orientation="horizontal")

        gui.separator(left_box_2)

        ###### direction space
        angular_distribution_box = oasysgui.widgetBox(left_box_2, "Direction space (divergences)", addSpace=True,
                                                      orientation="vertical") #, height=260)

        oasysgui.lineEdit(angular_distribution_box, self, "direction_space_center_x",
                          "Center in X' [rad]", tooltip="direction_space_center_x", labelWidth=260, valueType=float,
                          orientation="horizontal")

        oasysgui.lineEdit(angular_distribution_box, self, "direction_space_center_z",
                          "Center in Z' [rad]", tooltip="direction_space_center_z", labelWidth=260, valueType=float,
                          orientation="horizontal")

        oasysgui.lineEdit(angular_distribution_box, self, "direction_space_width_x",
                          "Width in X' [rad]", tooltip="direction_space_width_x", labelWidth=260, valueType=float, orientation="horizontal")

        oasysgui.lineEdit(angular_distribution_box, self, "direction_space_width_z",
                          "Width in Z' [rad]", tooltip="direction_space_width_z", labelWidth=260, valueType=float, orientation="horizontal")



        gui.separator(left_box_2)


        ##############################
        # ENERGY

        left_box_3 = oasysgui.widgetBox(tab_energy, "", addSpace=False, orientation="vertical") #, height=640)

        energy_wavelength_box = oasysgui.widgetBox(left_box_3, "Energy/Wavelength", addSpace=False,
                                                   orientation="vertical")

        gui.comboBox(energy_wavelength_box, self, "units", label="Units", labelWidth=260,
                     items=["Energy/eV", "Wavelength/Å"], orientation="horizontal")

        oasysgui.lineEdit(energy_wavelength_box, self, "single_line_value", "Value", tooltip="single_line_value", labelWidth=260, valueType=float,
                          orientation="horizontal")


        polarization_box = oasysgui.widgetBox(left_box_3, "Polarization", addSpace=False, orientation="vertical")

        self.ewp_box_8 = oasysgui.widgetBox(polarization_box, "", addSpace=False, orientation="vertical")
        oasysgui.lineEdit(self.ewp_box_8, self, "polarization_degree", "Polarization Degree [cos_s/(cos_s+sin_s)]",
                          tooltip="polarization_degree", labelWidth=310, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.ewp_box_8, self, "polarization_phase_deg", "Phase Difference [deg,0=linear,+90=ell/right]",
                          tooltip="polarization_phase_deg", labelWidth=310, valueType=float, orientation="horizontal")
        gui.comboBox(self.ewp_box_8, self, "coherent_beam", label="Phase of the sigma field", labelWidth=310,
                     tooltip="coherent_beam",items=["Random (incoherent)", "Constant (coherent)"], orientation="horizontal")

        gui.rubber(self.controlArea)
        gui.rubber(self.mainArea)

        self.set_coordinates_visibility()

    def is_scanning_enabled(self):
        return True

    def set_coordinates_visibility(self):
        self.points_cartesian_box.setVisible(self.coordinates == 0)
        self.points_polar_box.setVisible(self.coordinates == 1)

    def get_lightsource(self):
        import scipy.constants as codata
        if self.units == 0:
            wavelength = codata.h * codata.c / codata.e / self.single_line_value
        else:
            wavelength = self.single_line_value * 1e-10

        if self.coordinates == 0:
            gs =  SourceGridCartesian(
                real_space_width = [self.real_space_width_x, 0.0, self.real_space_width_z],
                real_space_center=[self.real_space_center_x, 0.0, self.real_space_center_z],
                real_space_points=[self.real_space_points_x, 1, self.real_space_points_z],
                direction_space_width = [self.direction_space_width_x, self.direction_space_width_z],
                direction_space_center = [self.direction_space_center_x, self.direction_space_center_z],
                direction_space_points=[self.direction_space_points_x, self.direction_space_points_z],
                wavelength=wavelength,
                polarization_degree=self.polarization_degree,
                polarization_phase_deg=self.polarization_phase_deg,
                coherent_beam=self.coherent_beam,
                name = "Grid Source (Cartesian)")
        else:
            gs =  SourceGridPolar(
                real_space_width = [self.real_space_width_x, 0.0, self.real_space_width_z],
                real_space_center=[self.real_space_center_x, 0.0, self.real_space_center_z],
                real_space_points = [self.real_space_points_r, self.real_space_points_theta],
                direction_space_width = [self.direction_space_width_x, self.direction_space_width_z],
                direction_space_center=[self.direction_space_center_x, self.direction_space_center_z],
                direction_space_points = [self.direction_space_points_r, self.direction_space_points_theta],
                wavelength=wavelength,
                polarization_degree=self.polarization_degree,
                polarization_phase_deg=self.polarization_phase_deg,
                coherent_beam=self.coherent_beam,
                name = "Grid Source (Polar)")

        return gs

    @Inputs.trigger
    def set_trigger_parameters_for_optics(self, trigger):
        super(OWGrid, self).set_trigger_parameters_for_optics(trigger)

    def run_shadow4(self, scanning_data: ShadowData.ScanningData = None):
        try:
            set_verbose()
            self.shadow_output.setText("")
            sys.stdout = EmittingStream(textWritten=self._write_stdout)

            self._set_plot_quality()

            self.progressBarInit()

            light_source = self.get_lightsource()

            self.progressBarSet(5)

            # run shadow4

            output_beam = light_source.get_beam()

            #
            # beam plots
            #
            self._plot_results(output_beam, None, progressBarValue=80)

            #
            # script
            #
            script = light_source.to_python_code()

            script += "\n\n# test plot\nfrom srxraylib.plot.gol import plot_scatter"
            script += "\nrays = beam.get_rays()"
            script += "\nplot_scatter(1e6 * rays[:, 0], 1e6 * rays[:, 2], title='(X,Z) in microns')"

            self.shadow4_script.set_code(script)

            self.progressBarFinished()

            #
            # send beam and trigger
            #
            output_data = ShadowData(beam=output_beam,
                                     number_of_rays=output_beam.get_number_of_rays(),
                                     beamline=S4Beamline(light_source=light_source))
            output_data.scanning_data = scanning_data

            self.Outputs.shadow_data.send(output_data)
            self.Outputs.trigger.send(TriggerIn(new_object=True))
        except Exception as exception:
            try:    self._initialize_tabs()
            except: pass
            self.prompt_exception(exception)

add_widget_parameters_to_module(__name__)
